/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')
let mg = require('mongoose')

let log = require('../../log').m('Bundle')

const CoreObject = require('../core/abstract/object')
const MgBundle = require('./mgBundle')

const Collection = require('./collection')
const Node = require('../node/node')

class Bundle extends CoreObject
{
  constructor(parent,dbmodel=null)
  {
    super(parent)

    this._new = true
    this._saved = false

    this._name = ''
    this._code = ''
    this._desc = ''
    this._flags = []
    this._prefix = 'b_'
    this._fields = []

    this._schema = null
    this._collection = null
    this._nodeClass = null

    if (dbmodel) this.fromDb(dbmodel)

    this.onCompiled = new Rx.Subject()
  }

  collection()
  {
    return this._collection
  }

  name(val)
  {
    if (typeof val == 'string')
    {
      this._name = val
      this._saved = false
    }
    return this._name
  }

  code(val)
  {
    if (typeof val == 'string')
    {
      this._code = val
      this._saved = false
    }
    return this._code
  }

  desc(val)
  {
    if (typeof val == 'string')
    {
      this._desc = val
      this._saved = false
    }
    return this._desc
  }

  prefix(val)
  {
    if (typeof val == 'string')
    {
      this._prefix = val
      this._saved = false
    }
    return this._prefix
  }

  flags(val)
  {
    if (Array.isArray(val))
    {
      this._flags = val
      this._saved = false
    }
    return this._flags
  }

  hasFlag(key)
  {
    return this._flags.reduce((acc,el)=>{
      if (acc) return acc
      if (el==key) return true
      return false
    },false)
  }

  addFlag(key)
  {
    if (this.hasFlag(key)) return
    this._flags.push(key)
    this._saved = false
  }

  delFlag(key)
  {
    if (!this.hasFlag(key)) return
    this._flags = this._flags.filter(f=>f!=key)
    this._saved = false
  }

  fields(val)
  {
    if (Array.isArray(val))
    {
      this._fields = val
      this._saved = false
    }
    return this._fields
  }

  getField(code)
  {
    return this._fields.reduce((acc,el)=>{
      return el.code==code?el:acc
    },null)
  }

  hasField(code)
  {
    return this._fields.reduce((acc,el)=>{
      return el.code==code?true:acc
    },false)
  }

  addField(field)
  {
    if (this.hasField(field.code)) return
    this._fields.push(field)
    this._saved = false
  }

  delField(code)
  {
    if (!this.hasField(code)) return
    this._fields = this._fields.filter(f=>f.code!=code)
    this._saved = false
  }

  updField(field)
  {
    if (!this.hasField(field.code)) return
    for(let i=0;i<this._fields.length;i++)
      if (this._fields[i].code == field.code)
        this._fields[i] = field
    this._saved = false
  }

  /**/
  toDb()
  {
    return {
      name: this._name,
      code: this._code,
      desc: this._desc,
      prefix: this._prefix,
      flags: this._flags,
      fields: this._fields
    }
  }

  fromDb(db)
  {
    //console.log(db)
    let ddb =  db
    if (typeof db.toObject == 'function')
      ddb = db.toObject()
    this._name = ddb.name
    this._code = ddb.code
    this._desc = ddb.desc
    this._prefix = ddb.prefix
    this._flags = ddb.flags
    this._fields = ddb.fields
    this._new = false
    this._saved = true
  }
  /**/

  save(recompile=true,dropData=false)
  {
    return Rx.Observable.create(obs=>{
      if (this._saved)
        return this.obsEnd(obs,this)
      log.log('Saving bundle',this._code)
      Rx.from(MgBundle.findOneAndUpdate({code:this._code},this.toDb(),{upsert:true,new: true})).subscribe(sb=>{
        this.fromDb(sb)
        log.log('bundle saved',this.code())
        if (!recompile)
          return this.obsEnd(obs,this)
        log.log('recompiling bundle',this.code())
        this.compile(dropData).subscribe(()=>{
          log.log('bundle recompiled',this.code())
          obs.next(this)
          obs.complete()
        },err=>{
          log.err('recompiling failed')
          console.log(err)
          obs.error(err)
        })
      },err=>{
        log.err('bundle save failed')
        console.log(err)
        obs.error(err)
      })
    })
  }

  compile(dropData=false)
  {
    return Rx.Observable.create(obs=>{
      log.log('compiling',this._code)
      let schema = {}
      this._fields.forEach(f=>{
        let fschema = this.M('field').fieldSchema(f)
        if (this._fieldHasFlag(f,'index'))
          fschema.index = true
        if (this._fieldHasFlag(f,'required'))
        {
          fschema.index = true
          fschema.required = true
        }
        if (typeof f.def != 'undefined')
          fschema.default = f.def
        schema[f.code] = [fschema]
      })
      this.M('db').dropBundleTable(this.dbTable(),dropData).subscribe(()=>{
        let clsNode = null
        let clsFn = null
        if (this._collection)
        {
          clsNode = this._collection._nodeClass || null
          clsFn = this._collection._classFn || null
        }

        let collectionClass = this._core.cls('collection')
        this._collection = new collectionClass(this,schema)
        if (clsNode)
          this._collection.setNodeClass(clsNode)
        if (clsFn)
          this._collection.setNodeClassFn(clsFn)
        log.log('compiled',this._code)
        this.onCompiled.next(this)
        let next = Rx.of(1)
        if (dropData)
          next = this._collection.removeNodes({})
        next.subscribe(()=>{
          obs.next()
          obs.complete()
        },err=>{obs.error(err)})
      })
    })
  }

  drop()
  {
    return this.M('db').dropBundleTable(this.dbTable(),true).pipe(
      RxO.mergeMap(_=>{
        return Rx.from(MgBundle.findOneAndRemove({code:this._code}))
      })
    )
  }

  dbTable()
  {
    return this.prefix()+this.code()
  }

  setNodeClass(Cls)
  {
    if (!this._collection) return
    this._collection.setNodeClass(Cls)
  }

  _fieldHasFlag(field,flag)
  {
    return field.flags.reduce((acc,el)=>{
      if (el==flag) return true
      return acc
    },false)
  }

}

module.exports = Bundle
