/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')
let mg = require('mongoose')

let log = require('../../log').m('BundleManager')
let g18n = require('../../g18n')

const CoreModule = require('../core/abstract/module')
const MgBundle = require('./mgBundle')

const Bundle = require('./bundle')
const Collection = require('./collection')
const Node = require('../node/node')

class BundleManager extends CoreModule
{
  constructor(parent)
  {
    super(parent)
    this._bundles = {}

    this._core.registerClass('bundle',Bundle)
    this._core.registerClass('collection',Collection)
    this._core.registerClass('node',Node)
  }

  moduleInit()
  {
    return Rx.Observable.create(obs=>{
      log.log('initializing')
      Rx.from(MgBundle.find({})).subscribe(list=>{
        let $t = []
        list.forEach(b=>{
          this._bundles[b.code] = new Bundle(this,b)
          $t.push(this._bundles[b.code].compile())
        })
        Rx.concat(...$t).subscribe(coll=>{
          //log.success('loaded bundle collection',coll.code())
        },e=>{
          log.err('failed!')
          console.log(e)
          obs.error(e)
        },()=>{
          log.log('initialized')
          obs.next()
          obs.complete()
        })
      })
    })
  }

  bundle(code,createNew=false)
  {
    if (this.bundleExists(code)&&!createNew) return this._bundles[code]
    if (!createNew) return null
    let ret = new Bundle(this)
    ret.code(code)
    ret.onCompiled.subscribe(bb=>{
      this._bundles[code] = bb
      //this._bundleCompiled(bb)
    })
    return ret
  }

  bundleExists(code)
  {
    if (this._bundles[code]) return true
    return false
  }

  bundleTable(code)
  {
    let b = this.bundle(code)
    if (!b) return null
    return b.dbTable()
  }

  collection(code)
  {
    if (!this.bundleExists(code)) return null
    return this.bundle(code).collection()
  }

  dashboardRoutes(app)
  {
    return Rx.Observable.create(obs=>{

      let rtBundleList = {
        url: '/api/bundle/list',
        access: 'auth.dashboard',
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let ret = []
            for (let code in this._bundles)
            {
              let b = this._bundles[code]
              if (!b) return
              ret.push({
                name: b.name(),
                code: b.code(),
                desc: b.desc(),
                prefix: b.prefix(),
                flags: b.flags(),
                fields: b.fields()
              })
              //console.log(b)
            }
            //console.log(ret)
            this.obsEnd(obs,ret)
          })
        }
      }


      let rtBundleSave = {
        url: '/api/bundle/save',
        access: 'bundle.root',
        method: 'post',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let b = ctx.f('body')
            let next = this._newBundle(b.bundle)
            if (this.bundle(b.bundle.code))
              next = this._updateBundle(b.bundle)
            next.subscribe(res=>{
              this.obsEnd(obs,{success:res})
            })
          })
        }
      }

      let rtBundleFields = {
        url: '/api/bundle/fields/:code',
        access: 'auth.dashboard',
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let p = ctx.f('params')
            let code = p.code || null
            if (!code)
              return this.obsEnd(obs,[])
            let bundle = this.bundle(code)
            if (!bundle)
              return this.obsEnd(obs,[])
            let ret = bundle.fields()
            //console.log(ret)
            this.obsEnd(obs,ret)
          })
        }
      }


      let rtBundleDel = {
        url: '/api/bundle/del',
        access: 'bundle.root',
        method: 'post',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let b = ctx.f('body')
            if (!this.bundle(b.bundle.code))
              this.obsEnd(obs,{success:false,error:'Bundle not found'})
            this._deleteBundle(b.bundle).subscribe(res=>{
              this.obsEnd(obs,{success:res})
            })
          })
        }
      }



      this.obsEnd(obs,[
        rtBundleList,
        rtBundleSave,
        rtBundleDel,
        rtBundleFields
      ])
    })
  }


  _newBundle(info)
  {
    return Rx.Observable.create(obs=>{
      //console.log(['new bundle',info])
      let b = new Bundle(this,info)
      b.prefix('u_')
      b.save().subscribe(bb=>{
        this._bundles[b.code()] = b
        this.obsEnd(obs,{success:true})
      },err=>{
        log.err('new bundle compile failed')
        console.log(err)
        this.obsEnd(obs,{success:false,error:'Bundle not compiled'})
      })
    })
  }

  _updateBundle(info)
  {
    return Rx.Observable.create(obs=>{
      log.info('updating bundle',info.code)
      let b = this.bundle(info.code)
      let regenFields = []
      info.fields.forEach(ff=>{
        if (!b.hasField(ff.code)&&ff.conf.genquantity&&(ff.generator!=''||typeof ff.def != 'undefined'))
        {
          regenFields.push(ff.code)
        }
      })
      b.name(info.name)
      b.desc(info.desc)
      b.fields(info.fields)

      b.save().subscribe(sb=>{
        if (!regenFields.length)
          return this.obsEnd(obs,{success:true})
        //let $t = regenFields.map(code=>b.collection().regenerateField(code))
        //console.log(regenFields)
        log.info('regenerating collection',info.code)
        b.collection().regenerateFields(regenFields).subscribe(rres=>{
          log.success('collection updated')
          this.obsEnd(obs,rres)
        },err=>{
          log.err('regenerating field failed')
          console.log(err)
          this.obsEnd(obs,{success:false,error:'Collection not updated'})
        })
      },err=>{
        log.err('bundle saving error!')
        console.log(err)
        this.obsEnd(obs,{success:false,error:'Bundle not saved!'})
      })
    })
  }

  _deleteBundle(info,dropData=false)
  {
    return Rx.of(this.bundle(info.code)).pipe(
      RxO.mergeMap(bndl=>{
        if (!bndl)
          return Rx.of(true)
        return bndl.drop()
      }),
      RxO.map(_=>{
        this._bundles[info.code] = null
        return true
      })
    )
  }

  _update01to011()
  {
    return Rx.of(1).pipe(
      RxO.mergeMap(_=>Rx.concat(
        this.M('auth/role').roleSave('bundle','root',g18n.t('Bundles root'))
      ).pipe(RxO.last()))
    )
  }

  moduleUpdate(from,to)
  {
    return Rx.of(1).pipe(
      RxO.mergeMap(_=>{
        if (from=='0.1')
          return this._update01to011()
        return Rx.of(1)
      }),
      RxO.map(_=>to)
    )
  }


  moduleDefaults()
  {
    return Rx.Observable.create(obs=>{
      let $t = []
      
      $t.push(this.M('auth/role').roleSave('bundle','root',g18n.t('Bundles root')))

      Rx.concat(...$t).subscribe(_=>{},err=>this.obsErr(obs,err),_=>this.obsEnd(obs))
    })
  }


}

module.exports = BundleManager
