/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


/**
@fileOverview Abstract core module class file
*/
let Rx = require('rxjs')

const CoreObject = require('./object')

/**
Core Module class - base parent class for any module, with default implementation of module behaviour.
@class CoreModule
@extends CoreObject
*/
class CoreModule extends CoreObject
{

  /**
  @constructor
  @param {CoreObject} [parent=null] - parent object, to get core objects api
  */
  constructor(parent=null)
  {
    super(parent)
    /**
      @private
      @name CoreModule#_config - module config, stored at files/modules.json
    */
      //@property {any} _config - module config, saved to files/modules.json
    this._config = null
  }

  /**
  Default module install method. Should be reimplemented by childs, if any install procedure required.
  This method calls only on installation routine and should return module config object, if module requires hold configs at /config/modules.json
  This method calls before {@link CoreModule#moduleConfigure} and before {@link CoreModule#moduleDefaults}
  @return {any} config=null - default module config
  */
  moduleInstall()
  {
    let defaultConfig = null
    return Rx.of(defaultConfig)
  }

  moduleUninstall()
  {
    return Rx.of(null)
  }

  /**
  Default module data creation
  */
  moduleDefaults()
  {
    return Rx.of(null)
  }

  /**
  Module configuration routine
  On reimplementation developer should care to save config data to {@link CoreModule._config}
  @param {any} [conf] - config object, loaded from /config/modules.json on 'normal start' or returned by {@link CoreModule#moduleInstall} routine
  */
  moduleConfigure(conf)
  {
    return Rx.Observable.create(obs=>{
      this._config = conf
      this.obsEnd(obs)
    })
  }

  /**
  Public method to get module config on runtime
  @return {object} - module config object
  */
  moduleConfig()
  {
    return this._config
  }

  /**
  Module load routine, calls after module object was created on loading
  */
  moduleLoad()
  {
    return Rx.of(null)
  }

  /**
  Module init routine, calls on start routine after after all modules were loaded and configured
  */
  moduleInit()
  {
    return Rx.of(null)
  }

  /**
  Module start routine, calls on start routine after after all modules were initialized
  */
  moduleStart()
  {
    return Rx.of(null)
  }

  /**
  Module stop routine, calls on module stop
  */
  moduleStop()
  {
    return Rx.of(null)
  }

  /**
  Module update routine, calls on core start, or module manager update events
  @param {string} fromVersion - version, installed in system
  @param {string} toVersion - current module version, mainly
  @return {string} version, to save at /config/modules.json as installed current
  */
  moduleUpdate(fromVersion,toVersion)
  {
    return Rx.of(toVersion)
  }

}

module.exports = CoreModule
