/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')
let request = require('request')

let Threads = require('worker_threads')

let log = require('../../log').m('CryptoModule')
let g18n = require('../../g18n')

const CoreModule = require('../core/abstract/module')

//const EthToken = require('./models/ethtoken')

const CoinMarketCap = require('./cmc')
const CryptoLayer = require('./cl')

class CryptoModule extends CoreModule
{
  constructor(parent)
  {
    super(parent)
    this._up = false
    this._cmc = new CoinMarketCap(this)
    this._cl = new CryptoLayer(this)
    this._clOk = false
  }

  updateCLLoop()
  {
    if (!this._up) return
    log.log('updating CryptoLayer')
    this._cl.update().subscribe(num=>{
      log.success('CryptoLayer updated. repeat in 1h',num)
      setTimeout(()=>{
        this.updateCLLoop()
      },60*60*1000) //1h for free api key
    },err=>{
      log.err('CryptoLayer update failed',err)
      setTimeout(()=>{
        this.updateCLLoop()
      },60*60*1000) //1h for free api key
    })
  }


  updateCMCLoop()
  {
    if (!this._up) return
    log.log('updating CoinMarketCap')
    this._cmc.update().subscribe(num=>{
      log.success('CoinMarketCap updated. repeat in 1h',num)
      setTimeout(()=>{
        this.updateCMCLoop()
      },60*60*1000) //1h for dev
    },err=>{
      log.err('CoinMarketCap update failed',err)
      setTimeout(()=>{
        this.updateCMCLoop()
      },60*60*1000) //60min now for dev
    })
  }

  updateETHTokensLoop()
  {
    if (!this._up) return
    this.updateETHTokens().subscribe(()=>{
      this._core.module('bundle').collection('ethtokens').countNodes().subscribe(num=>{
        log.success('updated eth tokens. repeat in 12h. known tokens:',num)
        setTimeout(()=>{
          this.updateETHTokensLoop()
        },12*60*60*1000)
      })
    },err=>{
      log.err('eth tokens not updated! repeat in 1 minute')
      console.log(err)
      setTimeout(()=>{
        this.updateETHTokensLoop()
      },60*1000)
    })
  }

  updateETHTokens()
  {
    return Rx.Observable.create(obs=>{
      log.log('Start eth tokens update')
      request({url: 'https://raw.githubusercontent.com/kvhnuke/etherwallet/mercury/app/scripts/tokens/ethTokens.json'},(error,response,body)=>{
        if (error || response.statusCode !== 200)
        {
          log.err('update tokens failed')
          console.log(error)
          obs.error(error)
          return
        }
        let coll = this._core.module('bundle').collection('ethtokens')

    	  let tokens = JSON.parse(body)
        log.log('Impoted tokens',tokens.length)
        //let err = false
        let $t = tokens.map(token=>{
          return coll.updateNode({$or:[
            {address:token.address},
            {symbol:token.symbol}
          ]},token,{upsert:true,new:true})
        })

        Rx.concat(...$t).subscribe($token=>{
          log.log('eth token',$token.field('symbol').get())
        },err=>{
          obs.error(err)
        },()=>{
          obs.next()
          obs.complete()
        })
      })
    })
  }

  requestJsonRx(url)
  {
    return Rx.Observable.create(obs=>{
      request({url: url,json: true},(error,response,body)=>{
        if (!error && response.statusCode === 200)
        {
          obs.next(body)
          obs.complete()
          return
        }
        obs.error(error)
      })
    })
  }

  moduleStart()
  {
    return Rx.Observable.create(obs=>{
      log.log('starting')
      if (!Threads.isMainThread)
        return this.obsEnd(obs)
      this._up = true
      setTimeout(()=>{
        this._cl.updateCoins().subscribe(ok=>{
          this.updateETHTokensLoop()
          this.updateCMCLoop()
          this.updateCLLoop()
          this.obsEnd(obs)
        },err=>{
          log.err('error!',err)
          this.obsEnd(obs)
        })
      },0)
      log.log('started')
      this.obsEnd(obs)
    })
  }

  moduleStop()
  {
    return Rx.Observable.create(obs=>{
      log.log('stoping')
      this._up = false
      log.success('stoped')
      obs.next()
      obs.complete()
    })
  }

  /*
  moduleInit()
  {
    return Rx.Observable.create(obs=>{
      log.log('initializing')
      log.success('initialized')
      this.obsEnd(obs)
    })
  }
  /**/

  moduleInstall()
  {
    return Rx.Observable.create(obs=>{
      log.info('installing')
      let $t = []

      let bndlToken = this._core.module('bundle').bundle('ethtokens',true)
      bndlToken.name(g18n.t('ETH Tokens'))
      bndlToken.desc(g18n.t('Ethereum tokens'))
      bndlToken.flags(['system'])
      bndlToken.addField({name:g18n.t('Symbol'),code:'symbol',type:'string',flags:['system','required','unique','logic','template','title']})
      bndlToken.addField({name:g18n.t('Address'),code:'address',type:'string',flags:['system','required','logic','template']})
      bndlToken.addField({name:g18n.t('Decimal'),code:'decimal',type:'number',flags:['system','logic','template']})
      bndlToken.addField({name:g18n.t('Type'),code:'type',type:'string',flags:['system','logic','template']})
      
      log.info('saving bundle',bndlToken.code())
      $t.push(bndlToken.save(true,true))
      
      let bndlCoin = this._core.module('bundle').bundle('coin',true)
      bndlCoin.name(g18n.t('Coin'))
      bndlCoin.desc(g18n.t('Money coins'))
      bndlCoin.flags(['system'])
      bndlCoin.addField({name:g18n.t('Symbol'),code:'symbol',type:'string',flags:['system','required','unique','logic','template','title']})
      bndlCoin.addField({name:g18n.t('Name'),code:'name',type:'string',flags:['system','required','unique','logic','template','title']})
      bndlCoin.addField({name:g18n.t('Description'),code:'description',type:'string',flags:['system','logic','template'],conf:{ops:[{source:'manual',manual:true}]}})
      bndlCoin.addField({name:g18n.t('Type'),code:'type',type:'string',flags:['system','required','logic','template'],def:'fiat'})
      bndlCoin.addField({name:g18n.t('Platform'),code:'platform',type:'string',flags:['system','logic','template','title']})
      bndlCoin.addField({name:g18n.t('Decims'),code:'decims',type:'number',flags:['system','logic','template'],def:2})
      bndlCoin.addField({name:g18n.t('CMC Rank'),code:'cmcrank',type:'number',flags:['system','logic','template'],def:-1})
      bndlCoin.addField({name:g18n.t('Last USD'),code:'lastusd',type:'number',conf:{ops:[{source:'manual',manual:true},{source:'manual',manual:2}]},flags:['system','logic','template'],def:1})
      bndlCoin.addField({name:g18n.t('USD change 1h'),code:'usdc1h',type:'number',conf:{ops:[{source:'manual',manual:true},{source:'manual',manual:2}]},flags:['system','logic','template'],def:0})
      bndlCoin.addField({name:g18n.t('USD change 1d'),code:'usdc1d',type:'number',conf:{ops:[{source:'manual',manual:true},{source:'manual',manual:2}]},flags:['system','logic','template'],def:0})
      bndlCoin.addField({name:g18n.t('USD change 1w'),code:'usdc1w',type:'number',conf:{ops:[{source:'manual',manual:true},{source:'manual',manual:2}]},flags:['system','logic','template'],def:0})

      log.info('saving bundle',bndlCoin.code())
      $t.push(bndlCoin.save(true,true))

      log.info('set config ','Coinmarketcap API_KEY')
      $t.push(this.M('config').setConf('cmc.key','','Coinmartketcap API_KEY'))
      
      log.info('set config ','Currencylayer API_KEY')
      $t.push(this.M('config').setConf('cl.key','','Currencylayer.com API_KEY'))
      
      Rx.concat(...$t).subscribe(()=>{},err=>{
        log.err('installing crypto failed')
        console.log(err)
        obs.error(err)
      },()=>{
        log.success('installed')
        obs.next()
        obs.complete()
      })
    })
  }

  moduleDefaults()
  {
    return Rx.Observable.create(obs=>{
      let install = {}
      let $t = []
      $t.push(this.M('core').consolePrompt(g18n.t('CurrencyLayer api key []:'),{default:''}).pipe(RxO.map(val=>{install['cl.key']=val})))
      $t.push(this.M('core').consolePrompt(g18n.t('CoinMarketCap api key []:'),{default:''}).pipe(RxO.map(val=>{install['cmc.key']=val})))
      $t.push(this.M('core').consolePrompt(g18n.t('Etherscan api key []:'),{default:''}).pipe(RxO.map(val=>{install['etherscan.key']=val})))

      Rx.concat(...$t).subscribe(()=>{},err=>this.obsErr(obs,err),_=>{
        let $tt = []

        $tt.push(this.M('config').setConf('cl.key',install['cl.key'],'Currencylayer api key'))
        $tt.push(this.M('config').setConf('cmc.key',install['cmc.key'],'Coinmarketcap api key'))
        $tt.push(this.M('config').setConf('etherscan.key',install['etherscan.key'],'Etherscan api key'))
        Rx.concat(...$tt).subscribe(()=>{},err=>this.obsErr(obs,err),__=>{
          this.obsEnd(obs)
        })
      })

    })
  }


}

module.exports = CryptoModule
