/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')
let Threads = require('worker_threads')

const http = require('http')
const express = require('express')
const WebSocket = require('ws')

let expressSession = require('express-session')
let RedisStore = require('connect-redis')(expressSession)
let expressFileUpload = require('express-fileupload')

let log = require('../../log').m('DashboardModule')
let g18n = require('../../g18n')

const DashboardRouter = require('./router')

const CoreModule = require('../core/abstract/module')

class DashboardModule extends CoreModule
{
  constructor(parent)
  {
    super(parent)
    this._server = null
    this._app = null
    this._ws = null
    this._store = null
    this._sess = null

    this._router = new DashboardRouter(this)
  }

  moduleStart()
  {
    return Rx.Observable.create(obs=>{
      if (!Threads.isMainThread)
      {
        log.err('dashboard starts only in main thread')
        return this.obsEnd(obs)
      }
      log.log('starting')
      let cfg = {
        host: this._config.host,
        port: this._config.port
      }
      this._server.listen(cfg)
      log.success('started at',cfg)
      this.obsEnd(obs)
    })
  }

  moduleStop()
  {
    return Rx.Observable.create(obs=>{
      log.log('stoping')
      log.success('stoped')
      obs.next()
      obs.complete()
    })
  }

  moduleInit()
  {
    return Rx.Observable.create(obs=>{
      log.log('initializing')
      this._conf = this._core.config

      this._app = express()
      this._server = http.Server(this._app)
      this._store = new RedisStore(this.M('db').moduleConfig().redis)
      this._sess = expressSession({
        store: this._store,
        secret: this._config.sessionSecret,
        resave: true,
        saveUninitialized: true
      })

      this._app.use(express.json({limit:'10mb',extended: true}))
      this._app.use(this._sess)
      this._app.use(express.static('public'))
      this._app.use('/files',express.static('files'))
      this._app.use('/docs',express.static('docs'))
      this._app.use('/emoji',express.static('public/assets/emoji/imgs'))
      this._app.use(expressFileUpload({ limits: { fileSize: this.M('file').moduleConfig().maxFileSize } }))
      /**/
      this._app.use((req,res,next)=>{
        let ctx = this.M('context').newContext(this,'dashboard_http')
        ctx.session(req.session)
        ctx.addField({
          name: g18n.t('URL'),
          code: 'url',
          type: 'string',
          quantity: 1,
          flags: ['virtual','readonly']
        })
        ctx.addField({
          name: g18n.t('Request method'),
          code: 'method',
          type: 'string',
          quantity: 1,
          flags: ['readonly']
        })
        ctx.f('url',req.url)
        ctx.f('method',req.method)
        let token = ctx.sfield('token')
        this._core.module('auth/access').dashboardTokenUser(token).subscribe(u=>{
          if (u) ctx.users([u])
          if (u)
            log.info('dashbord user',[req.url,u.titles()])
          else
            log.warn('dashboard guest',req.url)
          req._ctx = ctx
          next()
        },err=>{
          log.err('search user failed')
          console.log(err)
          res.status(401).location('/logout').end()
        })
      })
      this._server.on('listening',()=>{
        log.success('HTTP listen',{host:this._config.host,port:this._config.port})
      })
      this._router.init().subscribe(()=>{
        log.log('initialized')
        obs.next()
        obs.complete()
      })
    })
  }

  moduleInstall()
  {
    return Rx.Observable.create(obs=>{
      let install = {}
      let $t = []
      $t.push(this.M('core').consolePrompt(g18n.t('Dashboard host [0.0.0.0]:'),{default:'0.0.0.0'}).pipe(RxO.map(val=>{install.host=val})))
      $t.push(this.M('core').consolePrompt(g18n.t('Dashboard port [7076]:'),{default:'7076'}).pipe(RxO.map(val=>{install.port=val})))

      Rx.concat(...$t).subscribe(()=>{},err=>this.obsErr(obs,err),_=>{
        let cfg = {
          host: install.host,
          port: install.port,
          sessionSecret: 'asdqweasdzxcasdqweasdzxc'
        }

        this.obsEnd(obs,cfg)
      })

    })
  }





  dashboardRoutes(app)
  {
    return Rx.Observable.create(obs=>{
      let botId = {
        url: '/api/botid',
        access: null,
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            this.M('bot/screen').screensWithReaction('user_login').subscribe(screens=>{
              let cmds = []
              screens.forEach(screen=>{
                if(!screen.field('enabled').get()) return
                if (screen.field('action').get()!='text') return
                let ad = screen.field('actionData').get()
                if(!ad.text||!ad.text.hasCommand) return
                cmds.push(ad.text.command)
              })
              //console.log(this._core.config)
              let ver = this._core._pkg.version
              let proj = this.M('config').conf('project.name')
              let bot = this.M('config').conf('telegram.username')
              let locale = this.M('config').conf('locale.default')
              obs.next({
                version: ver,
                projectName: proj,
                username: bot,
                defLocale: locale,
                login: cmds[0]
              })
              obs.complete()
            })
          })
        }
      }

      obs.next([botId])
      obs.complete()
    })
  }

}

module.exports = DashboardModule
