/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')

let mg = require('mongoose')
mg.set('useCreateIndex', true)
mg.set('useFindAndModify', false)
mg.Promise = global.Promise

let log = require('../../log').m('DB')
let g18n = require('../../g18n')

const CoreModule = require('../core/abstract/module')
const ModuleManager = require('../module_manager/module')

class DB extends CoreModule
{
  constructor(parent)
  {
    super(parent)
    this._connection = null
  }

  moduleConfigure(conf)
  {
    return Rx.Observable.create(obs=>{
      this._config = conf
      let config = null
      if (this._config&&this._config.mongo)
        config = this._config.mongo
      log.test('configuring',config)
      if(!config||
          !config.host||
          !config.host.length||
          !config.db||
          !config.db.length)
      {
        log.err('config not correct')
        obs.error('DB_CONFIG_NOT_CORRECT')
        return
      }
      log.success('configured')
      this.obsEnd(obs)
    })
  }

  moduleInit()
  {
    return Rx.Observable.create(obs=>{
      if (this._connection)
        return this.obsEnd(obs)
      if(!this._config)
        return this.obsErr(obs,{code:'DB_NOT_CONFIGURED'})
      log.test('Connecting mongo db')
      let url = 'mongodb://'+this._config.mongo.host+'/'+this._config.mongo.db
      mg.connect(url,{ useNewUrlParser: true }).then(m=>{
        log.success('Mongo connected')
        this._connection = mg.connection
        this.obsEnd(obs)
      },e=>{
        this.obsErr(obs,{code:'MONGO_CONNECTION_ERROR',error:e,data:this._conf})
      })
    })
  }

  clearDB()
  {
    return Rx.Observable.create(obs=>{
      this._connection.db.dropDatabase(()=>{
        for(let m in this._connection.models)
          if (m[1]=='_')
            delete this._connection.models[m]
        this.obsEnd(obs)
      })
    })
  }

  dropBundleTable(tbl,dropData=false)
  {
    return Rx.Observable.create(obs=>{
      log.log('dropBundleTable',{table:tbl,withData:dropData})
      if (!this._connection||!this._connection.models[tbl])
        return this.obsEnd(obs)
      if (dropData)
      {
        //console.log(this._connection.models[tbl])
        this._connection.db.dropCollection(tbl,(err,res)=>{
          this._connection.models[tbl].collection.drop((err,res)=>{
            log.warn('data dropped!',tbl)
            delete this._connection.models[tbl]
            this.obsEnd(obs)
          })
        })
      } else {
        delete this._connection.models[tbl]
        this.obsEnd(obs)
      }
    })
  }

  /*
  moduleInstall()
  {
    return Rx.concat(
      this.moduleConfigure(),
      this.moduleInit()
    ).pipe(RxO.last())
  }
  */

  moduleInstall()
  {
    return Rx.Observable.create(obs=>{
      let install = {}
      let $t = []
      $t.push(this.M('core').consolePrompt(g18n.t('MongoDB host [localhost]:'),{default:'localhost'}).pipe(RxO.map(val=>{install['mongo.host']=val})))
      $t.push(this.M('core').consolePrompt(g18n.t('MongoDB database [takatanjs]:'),{default:'takatanjs'}).pipe(RxO.map(val=>{install['mongo.db']=val})))

      $t.push(this.M('core').consolePrompt(g18n.t('RedisDB host [localhost]:'),{default:'localhost'}).pipe(RxO.map(val=>{install['redis.host']=val})))
      $t.push(this.M('core').consolePrompt(g18n.t('RedisDB port [6379]:'),{default:'6379'}).pipe(RxO.map(val=>{
        let port = parseInt(val) || 6379
        install['redis.port']=port
      })))

      $t.push(this.M('core').consolePrompt(g18n.t('RedisDB dashboard db [1]:'),{default:'1'}).pipe(RxO.map(val=>{
        let db = parseInt(val) || 1
        install['redis.dashboard']=db
      })))
      //$t.push(this.M('core').consolePrompt(g18n.t('RedisDB http db [2]:'),{default:'2'}).pipe(RxO.map(val=>{
        //let db = parseInt(val) || 2
        //install['redis.http']=db
      //})))

      Rx.concat(...$t).subscribe(()=>{},err=>this.obsErr(obs,err),_=>{
        let cfg = {
          mongo: {
            host: install['mongo.host'],
            db: install['mongo.db']
          },
          redis: {
            host: install['redis.host'],
            port: install['redis.port'],
            db: install['redis.dashboard']
          //},
          //httpredis: {
            //host: install['redis.host'],
            //port: install['redis.port'],
            //db: install['redis.http']
          }
        }

        let $tt = []

        $tt.push(this.moduleConfigure(cfg))
        $tt.push(this.moduleInit())
        $tt.push(this.clearDB())

        Rx.concat(...$tt).subscribe(()=>{},err=>this.obsErr(obs,err),__=>{
          this.obsEnd(obs,cfg)
        })
      })

    })
  }



}

module.exports = DB
