/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxOps = require('rxjs/operators')
let Threads = require('worker_threads')

let log = require('../../log').m('EventsModule')
let g18n = require('../../g18n')

const CoreModule = require('../core/abstract/module')

const PostEventAction = require('./postevent.action')

class EventsModule extends CoreModule
{

  constructor(...args)
  {
    super(...args)

    this._etypes = {}

    this.onEvent = new Rx.Subject()

    this._workers = []

    if (!Threads.isMainThread)
    {
      Threads.parentPort.on('message',msg=>{
        //log.info('parent msg',msg)
        this.onEvent.next(msg)
      })
    }
  }

  eventType(code)
  {
    return this._etypes[code]
  }

  eventsTypes()
  {
    return Object.keys(this._etypes).map(x=>this._etypes[x])
  }

  /*
    inf = {
      code: 'event type unique code',
      name: g18n.t('Human readable event name'),
      data: [{
        ///data field description
      }],
      hasUser: true/false //one of data field called 'user' and has related user information
    }
  */
  registerEventType(inf)
  {
    if (!Array.isArray(inf.data))
      inf.data = []
    inf.hasUser = inf.data.reduce((acc,x)=>{
      return x.type == 'relation'&&x.conf&&Array.isArray(x.conf.ops)&&x.conf.ops[0]&&x.conf.ops[0].manual=='user'?true:acc
    },false)
    this._etypes[inf.code] = inf
  }

  addWorker(w)
  {
    let ind = w.threadId
    let ww = {
      run: false,
      worker: w,
      i: ind
    }
    ww.worker
      .on('message',msg=>{
        //this.onThreadMessage.next(msg)
        this.onEvent.next(msg)
        this._workers.filter(x=>x.run&&x.i!=ind).forEach(www=>{
          www.worker.postMessage(msg)
        })
      })
      .on('exit',res=>{
        log.warn('worker exited. removing',ind)
        this._workers = this._workers.filter(x=>x.i!=ind)
      })
      .on('online',res=>{
        ww.run = true
      })
    this._workers.push(ww)
  }

  pushEvent(e,coreOnly=false)
  {
    this.onEvent.next(e)
    if (Threads.isMainThread)
    {
      e.sender = 'core'
      this._workers.filter(x=>x.run).forEach(www=>{
        www.worker.postMessage(e)
      })
    } else {
      if (!coreOnly)
      {
        e.sender = Threads.workerData.iface
        Threads.parentPort.postMessage(e)
      }
    }
    //log.info('event pushed',e)
  }

  moduleInit()
  {
    return Rx.Observable.create(obs=>{

      this.registerEventType({
        code: 'core:start',
        name: g18n.t('Core started (shoot once)'),
        data: []
      })

      this.registerEventType({
        code: 'event:custom:user',
        name: g18n.t('Custom user event'),
        data: [{
          code: 'code',
          name: g18n.t('Event code'),
          type: 'string'
        },{
          code: 'user',
          name: g18n.t('User'),
          type: 'relation',
          conf: {ops:[{source:'manual',manual:'user'}]}
        },{
          code: 'line',
          name: g18n.t('String data'),
          type: 'string'
        },{
          code: 'num',
          name: g18n.t('Number data'),
          type: 'number'
        },{
          code: 'flag',
          name: g18n.t('On/off flag'),
          type: 'boolean'
        },{
          code: 'node',
          name: g18n.t('Any node'),
          type: 'relation'
        }]
      })

      this.registerEventType({
        code: 'event:custom:guest',
        name: g18n.t('Custom guest event'),
        data: [{
          code: 'code',
          name: g18n.t('Event code'),
          type: 'string'
        },{
          code: 'line',
          name: g18n.t('String data'),
          type: 'string'
        },{
          code: 'num',
          name: g18n.t('Number data'),
          type: 'number'
        },{
          code: 'flag',
          name: g18n.t('On/off flag'),
          type: 'boolean'
        },{
          code: 'node',
          name: g18n.t('Any node'),
          type: 'relation'
        }]
      })

      this.M('actions').registerAction(PostEventAction)
      this.M('bot/reactions').registerReactionType(PostEventAction)
      this.obsEnd(obs)
    })
  }



  dashboardRoutes(app)
  {
    return Rx.Observable.create(obs=>{

      let apiList = {
        url: '/api/events/ae/list',
        access: ['auth.dashboard'],
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let ret = {
              actions: this.M('actions').actionsTypes(),
              events: this.eventsTypes()
            }
            this.obsEnd(obs,ret)
          })
        }
      }

      this.obsEnd(obs,[apiList])
    })
  }



}


module.exports = EventsModule
