/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let fs = require('fs')
let Rx = require('rxjs')
let RxO = require('rxjs/operators')

let log = require('../../../log').m('DefaultThemeEngine')
let g18n = require('../../../g18n')

const pug = require('pug')

//const AbstractHttpEngine = require('./abstract')
const CoreObject = require('../../core/abstract/object')

class DefaultEngine extends CoreObject
{

  static info()
  {
    return {
      code: 'default',
      name: g18n.t('Default theme engine'),
      desc: g18n.t('Uses pug and liquidjs')
    }
  }

  constructor(...args)
  {
    super(...args)
    this._paths = {
      core: [],
      user: []
    }
  }

  _tplDir(p)
  {
    let inf = this.constructor.info()
    let ret = null
    if (p.type=='module')
    {
      ret = './'
      if (p.scope == 'core')
        ret += 'core/modules/'
      else
        ret += 'modules/'
      ret += p.path+'/tpl/'+inf.code
    }
    if (p.type=='theme')
    {
      ret = './themes/'+p.path+'/tpl'
    }
    return ret
  }

  _tplPath(tpl,ctx)
  {
    let ret = null
    let paths = ctx.vnode('theme').tplPaths()
    this._paths.user.forEach(p=>{
      paths.push(p)
    })
    this._paths.core.forEach(p=>{
      paths.push(p)
    })
    //log.warn('search template',tpl)
    //console.log(paths)
    paths.forEach(p=>{
      if (ret) return
      let path = this._tplDir(p)+'/'+tpl+'.pug'
      let stat = null
      try {
        stat = fs.statSync(path)
      } catch (e) {
        return
      }
      if (!stat||typeof(stat.isFile)!='function')
        return
      if (stat.isFile())
        ret = path
    })
    return ret
  }

  _tpl(names,ctx)
  {
    let srch = Array.isArray(names)?names:[names]
    return srch.reduce((acc,tpl)=>{
      if (acc) return acc
      let path = this._tplPath(tpl,ctx)
      if (path) return path
      return acc
    },null)
  }

  init(ctx)
  {
    //return Rx.Observable.create(obs=>{
    return this._core.hookCollect('hookHttpTemplatesPath').pipe(
      RxO.mergeMap(paths=>{
        log.log('initializing')
        //console.log(['collected templates paths',paths])
        return Rx.of(...paths)
      }),
      RxO.filter(p=>{
        let path = this._tplDir(p)
        //log.test('is tpl dir?',path)
        let stat = null
        try {
          stat = fs.statSync(path)
        } catch(e) {
          return false
        }
        if (!stat&&typeof(stat.isDirectory)!='function')
          return false
        if (stat.isDirectory())
        {
          log.success('yes',path)
          return true
        }
        return false
      }),
      RxO.reduce((acc,p)=>{
        this._paths[p.scope].push(p)
        acc.push(p)
        return acc
      },[])
    )
  }

  build(ctx)
  {
    return Rx.Observable.create(obs=>{
      let regions = ctx.vnode('theme').f('regions')
      //console.log(regions)
      let html = {
        blocks: [],
        regions: {}
      }
      let bids = []
      Object.keys(regions).forEach(rcode=>{
        html.regions[rcode] = []
        regions[rcode].forEach(bid=>{
          if (bids.indexOf(bid)==-1)
            bids.push(bid)
        })
      })
      if (!bids.length)
        return this.obsEnd(obs,html)
      this.C('http_blocks').findNodes({_id:{$in:bids}},{list:false}).pipe(
        RxO.mergeMap(b=>b.test(ctx).pipe(
          RxO.map(tr=>tr?b:null),
          RxO.filter(x=>!!x)
        )),
        RxO.reduce((acc,b)=>{
          acc.push(b)
          return acc
        },[])
      ).subscribe(blocks=>{
        html.blocks = blocks
        this.obsEnd(obs,html)
      },err=>this.obsErr(obs,err))
    })
  }

  compile(data,ctx)
  {
    return Rx.of(ctx.vnode('theme')).pipe(
      //rendering blocks
      RxO.mergeMap(theme=>{
        let regions = theme.f('regions')
        //console.log(['blocks compile',regions])
        let $t = data.blocks.map(b=>b.build(ctx).pipe(RxO.map(x=>{
          return {
            bid: b.id(),
            bld: x
          }
        })))
        return Rx.concat(...$t).pipe(RxO.reduce((acc,bld)=>{
          //console.log(['block compiled',bld])
          acc[bld.bid] = bld.bld
          return acc
        },{}))
      }),
      //render regions
      RxO.mergeMap(blocksHtml=>{
        let regions = ctx.vnode('theme').f('regions')
        let $t = Object.keys(regions).map(rcode=>{
          if (!data.regions[rcode]||!regions[rcode].length)
            return null
          let rdata = {
            code: rcode,
            blocks: []
          }
          regions[rcode].forEach(bid=>{
            if (!blocksHtml[bid])
              return
            rdata.blocks.push(blocksHtml[bid])
          })
          if (!rdata.blocks.length)
            return null
          return rdata
        }).filter(x=>!!x)
        //console.log(['regions compiled',$t])
        return Rx.of($t)
      }),
      /**
      //renderPage
      RxO.mergeMap(regionsHtml=>{
        //console.log(['regions renders',regionsHtml])
        return ctx.vnode('page').buildIndex({body:regionsHtml},ctx).pipe(RxO.map(x=>{
          console.log(['index builded',x])
          return x
        }))
      }),
      /**
      //renderIndex
      RxO.map(pageHtml=>{
          return {
            json:pageHtml
          }
      })
      /**/
 


    )
  }

  render(data,ctx)
  {
    return Rx.of(ctx.vnode('theme')).pipe(
      //rendering blocks
      RxO.mergeMap(theme=>{
        let regions = theme.f('regions')
        let $t = data.blocks.map(b=>this.renderBlock(b,ctx).pipe(RxO.map(x=>{
          return {
            bid: b.id(),
            bld: x
          }
        })))
        return Rx.concat(...$t).pipe(RxO.reduce((acc,bld)=>{
          acc[bld.bid] = bld.bld
          return acc
        },{}))
      }),
      //render regions
      RxO.mergeMap(blocksHtml=>{
        //console.log(['blocks renders',blocksHtml])
        let regions = ctx.vnode('theme').f('regions')
        let $t = Object.keys(regions).map(rcode=>{
          if (!data.regions[rcode]||!regions[rcode].length)
            return Rx.of(null)
          let rdata = {
            code: rcode,
            blocks: []
          }
          regions[rcode].forEach(bid=>{
            if (!blocksHtml[bid])
              return
            rdata.blocks.push(blocksHtml[bid])
          })
          return this.renderRegion(rdata,ctx).pipe(
            RxO.filter(x=>x&&x.length),
            RxO.map(x=>{
              return {
                region: rcode,
                html: x
              }
            })
          )
        })
        return Rx.concat(...$t).pipe(
          RxO.filter(x=>!!x),
          RxO.reduce((acc,x)=>{
            //console.log(['regions reduce',x,acc])
            acc[x.region] = x.html
            return acc
          },{})
        )
      }),
      //renderPage
      RxO.mergeMap(regionsHtml=>{
        //console.log(['regions renders',regionsHtml])
        return this.renderPage(regionsHtml,ctx).pipe(RxO.map(x=>{
          return {
            body:x
          }
        }))
      }),
      //renderIndex
      RxO.mergeMap(pageHtml=>{
        return this.renderIndex(pageHtml,ctx).pipe(RxO.map(x=>{
          //console.log(['page renders',pageHtml,x])
          return {
            html:x
          }
        }))
      })
    )
  }

  renderBlock(block,ctx)
  {
    return block.build(ctx).pipe(
      RxO.mergeMap(bld=>{
        let code = block.f('code')
        let type = block.f('type')
        let tpl = this._tpl([
          'block-'+code,
          'block-'+type,
          'block'],ctx)
        if (!tpl)
          return Rx.of('')

        let rndr = pug.renderFile(tpl,{block:bld})

        //console.log(['BLOCK TPL!',tpl,rndr])
        return Rx.of(rndr)
      })
    )
  }

  renderRegion(region,ctx)
  {
    return Rx.Observable.create(obs=>{
      let tpl = this._tpl([
        'region-'+region.code,
        'region'
      ],ctx)
      if (!tpl)
        return this.obsEnd(obs,'')
      let rndr = pug.renderFile(tpl,{region:region})
      //console.log(['REGION TPL',tpl,rndr])
      this.obsEnd(obs,rndr)
    })
  }

  renderPage(regionsHtml,ctx)
  {
    return ctx.vnode('page').build(regionsHtml,ctx).pipe(
      RxO.map(bld=>{
        //console.log(['page builded',bld])
        let lst = [
          'page-'+ctx.vnode('page').f('code'),
          'page-'+ctx.vnode('page').f('type'),
          'page'
        ]
        if (ctx.vnode('content'))
          lst.unshift('page-'+ctx.vnode('content').bundle())

        let tpl = this._tpl(lst,ctx)
        if (!tpl)
          return ''
        let rndr = ''
        try {
          rndr = pug.renderFile(tpl,{page:bld})
        } catch(e) {
          if (e&&typeof e.toString == 'function')
            rndr = e.toString()
          else
            rndr = JSON.stringify(e)
        }
        //console.log(['page rendered',bld,rndr])
        return rndr
      })
    )
  }

  renderIndex(pageHtml,ctx)
  {
    return Rx.of(ctx.vnode('page')).pipe(
      RxO.mergeMap(page=>{
        //console.log(page)
        return page.buildIndex(pageHtml,ctx)
      }),
    //return ctx.vnode('page').buildIndex(pageHtml,ctx).pipe(
      RxO.map(bld=>{
        //console.log(['index builded',bld])
        let tpl = this._tpl(['index'],ctx)
        if (!tpl)
          return ''
        let rndr = ''
        try {
          rndr = pug.renderFile(tpl,{page:bld})
        } catch(e) {
          if (e&&typeof e.toString == 'function')
            rndr = e.toString()
          else
            rndr = JSON.stringify(e)
        }
        //console.log(['index rendered',bld,rndr])
        return rndr
      })
    )
  }

  /*
  compileBlock(block,ctx)
  {
    return block.build(ctx).pipe(
      RxO.mergeMap(bld=>{
        let type = block.f('type')
        let code = block.f('code')
        let tpl = this._tpl([
          'block-'+type,
          'block-'+code,
          'block'],ctx)
        if (!tpl)
          return Rx.of('')

        let rndr = pug.renderFile(tpl,{block:bld})

        //console.log(['BLOCK TPL!',tpl,rndr])
        return Rx.of(rndr)
      })
    )
  }
  /**/


}

module.exports = DefaultEngine
