/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')

const http = require('http')
const express = require('express')

let expressSession = require('express-session')
let RedisStore = require('connect-redis')(expressSession)
let expressFileUpload = require('express-fileupload')
let expressBodyParser = require('body-parser')

let log = require('../../log').m('HttpInterface')
let g18n = require('../../g18n')

const AbstractInterface = require('../interface/abstract')
const VNode = require('../node/vnode')

class HttpInterface extends AbstractInterface
{

  constructor(parent,model,isNew=false)
  {
    super(parent,model,isNew)
    this._type = 'iface'

    this._server = null
    this._app = null
    this._store = null
    this._sess = null
  }

  static ifInfo()
  {
    return {
      code: 'http',
      name: g18n.t('Http'),
      desc: g18n.t('Http interface.'),
      scope: 'http',
      directions: 1,
      confs: [
        { code: 'host', name: g18n.t('Host'), desc: g18n.t('Listen host'), type: 'string', def: '0.0.0.0' },
        { code: 'port', name: g18n.t('Port'), desc: g18n.t('Listen port'), type: 'number', def: 8080 },
        { code: 'redisHost', name: g18n.t('Redis host'), desc: g18n.t('Redis db host for session storage'), type: 'string', def: '127.0.0.1' },
        { code: 'redisPort', name: g18n.t('Redis port'), desc: g18n.t('Redis db port for session storage'), type: 'number', def: 6379 },
        { code: 'redisDB', name: g18n.t('Redis DB'), desc: g18n.t('Redis db number for session storage'), type: 'number', def: 2 },
        { code: 'secret', name: g18n.t('Session secret'), desc: g18n.t('Unique string to encrypt session'), type: 'string', def: 'session-secret-code' },
        { code: 'publicDirs', name: g18n.t('Public folders'), desc: g18n.t('Folders data will be public accessible from /'), type: 'string', def: 'public', quantity: 0 },
        { code: 'uploadDir', name: g18n.t('Upload directory'), desc: g18n.t('Folder to upload files'), type: 'string', def: 'public' },
        { code: 'uploadMaxSize', name: g18n.t('Upload max size'), desc: g18n.t('Maximum size for upload files (bytes)'), type: 'number', def: 52428800 },
        { code: 'jsonMaxSize', name: g18n.t('JSON max size'), desc: g18n.t('Maximum size of json data, mb'), type: 'string', def: '50mb' },
      ]
    }
  }

  ifContext(demo=false)
  {
    let ctx = this.M('context').newContext(this,'http',true)

    let http = new VNode(ctx)

    http.addField({name:g18n.t('Interface code'),code:'code',type:'string'})
    http.addField({name:g18n.t('Interface type'),code:'type',type:'string'})
    http.addField({name:g18n.t('Request url'),code:'url',type:'string'})
    http.addField({name:g18n.t('Request method'),code:'method',type:'string'})
    http.addField({name:g18n.t('Request params'),code:'params',type:'any'})
    http.addField({name:g18n.t('Request query'),code:'query',type:'any'})
    http.addField({name:g18n.t('Request body'),code:'body',type:'any'})
    http.addField({name:g18n.t('User'),code:'user',type:'relation',xtype:'relation:user'})
    //http.addField({code:'files',type:'any'})

    ctx.addNode(http,'http')
    ctx.addNode(this,'interface')

    if (!demo)
      return ctx

    let u = this.C('user').newNode()
    ctx.users([u])

    let route = this._rc().newNode()

    return ctx
  }

  ifInit(start=false)
  {
    return Rx.of(this.f('config')).pipe(
      RxO.map(cfg=>{
        //console.log(cfg)
        /**/
        this.ifStatus('initializing')
        this._app = express()
        this._server = http.Server(this._app)
        this._store = new RedisStore({
          host: cfg.redisHost,
          port: cfg.redisPort,
          db: cfg.redisDB
        })
        this._sess = expressSession({
          store: this._store,
          secret: cfg.secret,
          resave: true,
          saveUninitialized: true
        })

        this._app.use(expressBodyParser.urlencoded({ extended: false }))
        this._app.use(express.json({limit:cfg.jsonMaxSize,extended: true}))
        this._app.use(this._sess)
        //TODO currently public access to files
        this._app.use('/files/',express.static('files'))
        cfg.publicDirs.forEach(d=>{
          this._app.use('/',express.static(d))
        })

        this._app.use((req,res,next)=>{
          let ctx = this.ifContext()
          ctx.session(req.session)

          let http = ctx.vnode('http')
          http.f('url',req.url)
          http.f('method',req.method)
          http.f('params',req.params)
          http.f('query',req.query)
          http.f('body',req.body)
          http.f('code',this.f('code'))
          let inf = this.constructor.ifInfo()
          http.f('type',inf.code)

          let token = ctx.sfield('token')
          this._core.module('auth/access').findIfaceSession(this.f('code'),token).subscribe(u=>{
            if (u) {
              log.info('iface user',[this.f('code'),req.url,u.titles()])
            } else {
              u = this.C('user').newNode()
              u.isGuest = true
              log.warn('iface guest',[this.f('code'),req.url])
            }
            ctx.users([u])
            http.f('user',u)
            req._ctx = ctx
            next()
          },err=>{
            log.err('search user failed')
            console.log(err)
            req._ctx = ctx
            next()
          })
        })
 
        this._app.use(expressFileUpload({ limits: { fileSize: cfg.uploadMaxSize } }))

        this._app.all('*',(req,res)=>{
          let ctx = req._ctx
          log.info(this.f('code'),[req.method,req.url])
          let router = this.M('interface').ifRouter(this.f('router'))
          if (!router)
          {
            res.status(500).send('Error: NULL_ROUTER')
            return
          }
          ctx.addNode(router,'router')
          let theme = this.M('http/theme').theme(router.f('config').theme)
          let localFile = theme.hasLocalFile(req.url)
          if (localFile&&localFile.length)
          {   
            let dir = __dirname.split('/').slice(0,-3).join('/')
            if (localFile.endsWith('.vue'))
            {   
              log.success('vue file')
              res.header("Content-Type", "application/javascript; charset=utf-8");
              res.status(200);
              res.send(fs.readFileSync(localFile).toString())
              return
            }   
            log.success('local file',localFile)
            res.sendFile(localFile,{root:dir})
            return
          } 
          //console.log(['if search route'])
          router.searchRoute(ctx//).pipe(
            //RxO.mergeMap(rt=>router.executeContext(ctx))
          ).subscribe(data=>{
            //console.log(ctx.view())
            //console.log(['res',data])

            if (!data)
            {
              res.status(500).send('Data error: '+JSON.stringify(data))
              return
            }
            let code = data.code || 200
            if (ctx.vnode('route')&&ctx.vnode('route').f('rescode'))
              code = ctx.vnode('route').f('rescode')
            if (typeof data.redirect == 'string')
            {
              let rurl = data.redirect
              if (!rurl.startsWith('/')&&!rurl.startsWith('http://')&&!rurl.startsWith('https://'))
                rurl = '/'+rurl
              res.redirect(302,rurl)
              return
            }
            if (typeof data.html == 'string')
            {
              if (req.url.endsWith('.js'))
              {
                log.success('js result!')
                res.header("Content-Type", "application/javascript; charset=utf-8");
              } else {
                log.success('html')
              }
              res.status(code).send(data.html)
              return
            }
            if (typeof data.js == 'string')
            {
              log.success('js result!')
              res.header("Content-Type", "application/javascript; charset=utf-8");
              res.status(code).send(data.js)
              return
            }
            if (data.json)
            {
              res.status(code).json(data.json)
              return
            }
            res.status(500).send('Data format unknown: '+JSON.stringify(data))



            //res.json({opopop:11})
          },err=>{
            log.err('search route error')
            console.log(err)
            res.status(500).send('Error: SEARCH_ROUTE_ERROR')
          })
        })

        this._server.on('listening',()=>{
          log.success('HTTP listen',{host:cfg.host,port:cfg.port})
          this.ifRun(true)
          this.ifStatus('run')
        })

        this._server.on('close',err=>{
          log.warn('HTTP close',err)
          this.ifStatus('stoped')
        })

        this._server.on('error',err=>{
          log.warn('HTTP error',err)
        })
        return true
      }),
      RxO.mergeMap(initRes=>Rx.iif(_=>initRes&&start,this.ifStart(),Rx.of(true)))
    )
    //return Rx.throwError({code:'ABSTRACT_IFACE'})
  }

  ifAttach(iface)
  {
    return Rx.throwError({code:'HTTP_IFACE_IS_NOT_ATTACHABLE'})
  }

  ifStart()
  {
    return Rx.of(this.f('config')).pipe(
      RxO.map(cfg=>{
        if (this._run)
          return true
        if (this.ifStatus()=='stopping')
          return false
        this._server.listen({host:cfg.host,port:cfg.port})
        return true
      })
    )
  }

  ifStop()
  {
    return Rx.of(this.f('config')).pipe(
      RxO.map(cfg=>{
        if (!this._run)
          return true
        this.ifRun(false)
        this.ifStatus('stopping')
        this._server.close()
        return true
      })
    )
  }

}

module.exports = HttpInterface
