/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')
let fs = require('fs')
let DT = require('directory-tree')

let log = require('../../../log').m('HttpThemeModule')
let g18n = require('../../../g18n')

const CoreModule = require('../../core/abstract/module')

const Theme = require('./theme')

class HttpThemeModule extends CoreModule
{

  constructor(parent)
  {
    super(parent)
    this._themes = {}
  }

  scanThemes()
  {
    let dir = DT('./themes/')
    return dir.children.map(x=>{
      let isTheme = Array.isArray(x.children)&&x.children.reduce((acc,xx)=>xx.name=='theme.json'?true:acc,false)
      if (!isTheme) return null
      let json = null
      try
      {
        json = JSON.parse(fs.readFileSync('./themes/'+x.name+'/theme.json'))
      } catch(e) {
        log.err('reading theme.json error!',[x.name,e])
        json = null
      }
      if (!json) return null
      return {
        code: x.name,
        info: json
      }
    }).filter(x=>!!x)
  }

  dashboardRoutes(app)
  {
    return Rx.Observable.create(obs=>{

      let rtThemes = {
        url: '/api/http/themes',
        access: ['auth.dashboard','http.theme'],
        method: 'get',
        fn: (ctx)=>{
          return this.C('http_themes').findNodes().pipe(
            RxO.map(list=>{
              let all = this.scanThemes()
              let theme = this.M('config').conf('http.theme')
              return {
                all: all,
                themes: list.filter(x=>!!x).map(x=>x.toObject()),
                current: theme
              }
            })
          )
        }
      }

      let rtThemeSave = {
        url: '/api/http/theme/save',
        access: ['auth.dashboard','http.theme'],
        method: 'post',
        fn: (ctx)=>{
          return Rx.of(ctx.f('body')).pipe(
            RxO.mergeMap(data=>{
              if (!data||!data.ep||!data.data)
                return Rx.of({ error: 'DATA_NOT_CORRECT' })
              //console.log(['themeSave',data])
              let theme = this.theme(data.data.code)
              if (!theme)
                return Rx.of({ error: 'THEME_NOT_FOUND' })
              if (data.ep=='regions')
                theme.f('regions',data.data.regions)
              if (data.ep=='confs')
              {
                let cfg = theme.f('conf')
                //console.log(data.data,cfg)
                cfg.conf = data.data.conf.conf
                theme.f('conf',cfg)
              }
              return theme.save()
            }),
            RxO.map(_=>{
              return {success: true}
            })
          )
        }
      }

      let rtThemeEnable = {
        url: '/api/http/theme/enable/:code',
        access: ['auth.dashboard','http.theme'],
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let code = ctx.f('params').code || null
            this.enableTheme(code).subscribe(()=>{
              this.obsEnd(obs,{success:true})
            },err=>{
              this.obsEnd(obs,{success:false,error:err})
            })
          })
        }
      }

      let rtThemeDisable = {
        url: '/api/http/theme/disable/:code',
        access: ['auth.dashboard','http.theme'],
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let code = ctx.f('params').code || null
            this.disableTheme(code).subscribe(()=>{
              this.obsEnd(obs,{success:true})
            },err=>{
              this.obsEnd(obs,{success:false,error:err})
            })
          })
        }
      }

      let rtThemeUninstall = {
        url: '/api/http/theme/uninstall/:code',
        access: ['auth.dashboard','http.theme'],
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let code = ctx.f('params').code || null
            this.uninstallTheme(code).subscribe(()=>{
              this.obsEnd(obs,{success:true})
            },err=>{
              this.obsEnd(obs,{success:false,error:err})
            })
          })
        }
      }



      let rtThemeSetDefault = {
        url: '/api/http/theme/setdefault/:code',
        access: ['auth.dashboard','http.theme'],
        method: 'get',
        fn: (ctx)=>{
          return Rx.Observable.create(obs=>{
            let code = ctx.f('params').code || null
            this.setDefaultTheme(code).subscribe(()=>{
              this.obsEnd(obs,{success:true})
            },err=>{
              this.obsEnd(obs,{success:false,error:err})
            })
          })
        }
      }

      this.obsEnd(obs,[
        rtThemes,
        rtThemeSave,
        rtThemeEnable,
        rtThemeDisable,
        rtThemeUninstall,
        rtThemeSetDefault
      ])
    })
  }

  moduleInit()
  {
    return Rx.Observable.create(obs=>{
      this.C('http_themes').setNodeClassFn(mdl=>{
        if (!mdl||!mdl.code||!mdl.code[0])
          return Theme
        let code = mdl.code[0]
        let Cls = null
        try
        {
          Cls = require('../../../../themes/'+code+'/theme')
        } catch(e) {
          log.err('theme class import error!',e)
          return Theme
        }
        //console.log(['CLS',code,Cls])
        if (Cls&&Theme.isPrototypeOf(Cls))
          return Cls
        return Theme
      })

      this.C('http_themes').findNodes({enabled:true},{list:false}).pipe(
        RxO.mergeMap(x=>this._enableTheme(x)),
        RxO.last()
      ).subscribe(_=>{
        this.obsEnd(obs)
      },err=>this.obsErr(obs,err))
    })
  }

  moduleInstall()
  {
    return Rx.Observable.create(obs=>{
      let $t = []

      //themes data bundle
      let bndThemes = this.M('bundle').bundle('http_themes',true)
      bndThemes.name(g18n.t('Web Themes'))
      bndThemes.desc(g18n.t('Web site themes'))
      bndThemes.addFlag('system')
      bndThemes.addField({name:g18n.t('Code'),code:'code',type:'string',flags:['system','required','logic','template']})
      bndThemes.addField({name:g18n.t('Name'),code:'name',type:'string',flags:['system','required','logic','template']})
      bndThemes.addField({name:g18n.t('Enabled'),code:'enabled',type:'boolean',flags:['system','required','logic','template'],def:false})
      bndThemes.addField({name:g18n.t('Conf'),code:'conf',type:'any',flags:['template','logic']})
      bndThemes.addField({name:g18n.t('Base'),code:'base',type:'string',flags:['template','logic']})
      bndThemes.addField({name:g18n.t('Engine'),code:'engine',type:'string',flags:['template','logic']})
      bndThemes.addField({name:g18n.t('Regions'),code:'regions',type:'any',flags:['template','logic']})
      bndThemes.addField({name:g18n.t('Features'),code:'features',type:'string',quantity:0,flags:['template','logic']})

      log.log('installing bundle',bndThemes.code())
      $t.push(bndThemes.save(true,true))

      Rx.concat(...$t).subscribe(()=>{},obs.error,()=>{
        log.success('bundle installed')
        obs.next()
        obs.complete()
      })
 
    })
  }

  moduleDefaults()
  {
    return Rx.of(null).pipe(RxO.mergeMap(_=>{
      let $t = []
      $t.push(this.installTheme('blocks',true))
      $t.push(this.installTheme('takatan',true))
      $t.push(this.setDefaultTheme('takatan'))
      return Rx.concat(...$t).pipe(RxO.last())
    }))
  }

  baseTheme(code)
  {
    if (!this._themes[code])
      return null
    let base = this._themes[code].f('base')
    if (!this._themes[base])
      return null
    return this._themes[base]
  }

  theme(code='')
  {
    let theme = this.M('config').conf('http.theme')
    if (code!=''&&this._themes[code])
      return this._themes[code]
    return this._themes[theme]
  }

  currentTheme()
  {
    let code = this.M('config').conf('http.theme')
    return this._themes[code]
  }

  installTheme(code,enabled=false)
  {
    let json = null
    try
    {
      json = JSON.parse(fs.readFileSync('./themes/'+code+'/theme.json'))
    } catch(e) {
      return Rx.throwError({code:'THEME_JSON_READ_ERROR'})
    }
    if (!json||!json.name)
      return Rx.throwError({code:'THEME_JSON_NOT_CORRECT'})
    return Rx.forkJoin(
      this.C('http_blocks').findNode({code:'content'}),
      this.C('http_themes').findNode({code:code})
    ).pipe(
      RxO.mergeMap(data=>{
        let cblock = data[0]
        let theme = data[1]
        if (!theme)
          theme = this.C('http_themes').newNode()
        theme.f('code',code)
        theme.f('enabled',enabled)
        theme.f('name',json.name)
        theme.f('base',json.base)
        theme.f('engine',json.engine)
        let regions = {}
        json.regions.forEach(r=>{
          regions[r] = []
        })
        //console.log(['content block',regions.content,cblock])
        if (regions.content&&cblock)
          regions.content.push(cblock.id())
        theme.f('regions',regions)
        let conf = {
          css: [],
          js: [],
          conf: {}
        }
        if (Array.isArray(json.css))
          conf.css = json.css
        if (Array.isArray(json.js))
          conf.js = json.js
        if (Array.isArray(json.confs))
          json.confs.forEach(cfg=>{
            conf.conf[cfg.code] = cfg.def
          })
        theme.f('conf',conf)
        log.info('installing theme',code)
        return theme.save().pipe(RxO.mergeMap($th=>{
          if (enabled)
            return this._enableTheme($th)
          return Rx.of($th)
        }))
      })
    )
  }

  uninstallTheme(code)
  {
    return Rx.Observable.create(obs=>{
      this.obsEnd(obs)
    })
  }

  enableTheme(code)
  {
    return this.C('http_themes').findNode({code:code}).pipe(
      RxO.mergeMap(theme=>{
        if (!theme)
          return this.installTheme(code,true)
        return this._enableTheme(theme)
      })
    )
  }

  _enableTheme(theme)
  {
    let nx = Rx.of(theme)
    if (!theme.f('enabled'))
    {
      theme.f('enabled',true)
      nx = theme.save()
    }
    return nx.pipe(
      RxO.mergeMap(thm=>{
        this._themes[theme.f('code')] = thm
        return Rx.of(thm)
      })
    )
  }

  disableTheme(code)
  {
    return this.C('http_themes').findNode({code:code}).pipe(
      RxO.mergeMap(theme=>{
        if (!theme)
          return Rx.of(true)
        theme.f('enabled',false)

        return theme.save().pipe(RxO.map($th=>{
          delete this._themes[code]
          this._themes[code] = undefined
          return true
        }))
      })
    )
  }

  uninstallTheme(code)
  {
    return this.disableTheme(code).pipe(
      RxO.mergeMap(_=>this.C('http_themes').removeNodes({code:code}))
    )
  }


  setDefaultTheme(code)
  {
    return this.M('config').setConf('http.theme',code,'Web site theme')
  }


}

module.exports = HttpThemeModule
