/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')
let Threads = require('worker_threads')

let log = require('../../log').m('AbstractInterface')
let g18n = require('../../g18n')

const Node = require('../node/node')

class AbstractInterface extends Node
{

  constructor(...args)
  {
    super(...args)

    this._status = 'unknown'
    this._run = false
    this._worker = null

    this.statusChange = new Rx.Subject()
    this.runChange = new Rx.Subject()
    //used from parent thread
    this.onThreadStart = new Rx.Subject()
    this.onThreadStop = new Rx.Subject()
    this.onThreadError = new Rx.Subject()
    this.onThreadMessage = new Rx.Subject()
    //used from child thread
    this.onParentMessage = new Rx.Subject()
    if (!Threads.isMainThread&&!!Threads.parentPort)
      Threads.parentPort.on('message',msg=>this.onParentMessage.next(msg))
  }

  ifStatus(next)
  {
    if (typeof next == 'string'&&this._status!=next)
    {
      this._status = next
      this.statusChange.next(this._status)
    }
    return this._status
  }

  ifRun(next)
  {
    if (typeof next == 'boolean'&&this._run!=next)
    {
      this._run = next
      this.runChange.next(this._run)
    }
    return this._run
  }

  //common routines
  static ifInfo()
  {
    return {
      code: 'abstract',
      name: g18n.t('Abstract'),
      desc: g18n.t('Abstract interface'),
      //interface 'network'
      //examples:
      //http, websocket - http
      //tgbot, tguserbot - telegram
      //fbbot - facebook
      scope: 'abstract', 
      //1 if users request required to send something to user (like http)
      //2 if we can send something to user any moment
      directions: 0,
      confs: [],
      //interface type, required for attachment
      requires: ''
    }
  }

  ifContext(demo=false)
  {
    let ctx = this.M('context').newContext(this,'abstract')
    let iface = new VNode(ctx)
    iface.addField({name:g18n.t('Interface code'),code:'code',type:'string'})
    iface.addField({name:g18n.t('Interface type'),code:'type',type:'string'})
    ctx.addNode(iface,'interface')
    return ctx
  }

  ifReconfigure(oldCfg)
  {
    return this.stopIface().pipe(
      RxO.mergeMap(ok=>{
        if (!ok)
          return Rx.throwError({code:'IFSTOP_FAILED'})
        if (this.f('enabled'))
          return this.startIface(this.f('enabled'))
        return Rx.of(1)
      })
    )
    //return Rx.throwError({code:'ABSTRACT_IFACE'})
  }

  //used from child thread
  ifInit(start=false)
  {
    return Rx.throwError({code:'ABSTRACT_IFACE'})
  }

  ifAttach(iface)
  {
    return Rx.throwError({code:'ABSTRACT_IFACE'})
  }

  ifStart()
  {
    return Rx.throwError({code:'ABSTRACT_IFACE'})
  }

  ifStop()
  {
    return Rx.throwError({code:'ABSTRACT_IFACE'})
  }

  //used from any thread
  //magic happens exactly here

  startIface(run=false)
  {
    return Rx.iif(_=>Threads.isMainThread,
      Rx.of(1).pipe(
        RxO.tap(_=>log.warn('starting in thread',this.f('code'))),
        RxO.mergeMap(_=>{
          return this.startThread({
            mode:'iface',
            run: run,
            iface:this.f('code'),
            logCode: this.f('code')
          })
        })
      ),
      Rx.of(Threads.workerData).pipe(
        RxO.tap(_=>log.warn('starting as thread',this.f('code'))),
        RxO.mergeMap(wd=>{
          //console.log([wd,this.f('code')])
          if (wd.mode!='iface'||wd.iface!=this.f('code'))
          {
            log.err('this is not my start',this.f('code'))
            return Rx.of(true)
          }
          return this.ifInit(wd.run)
        })
      )
    )
  }

  stopIface()
  {
    return Rx.iif(_=>Threads.isMainThread,
      this.stopThread(),
      this.ifStop()
    )
  }

  //used from parent thread

  startThread(data)
  {
    return Rx.Observable.create(obs=>{
      log.warn('starting thread!')
      //starting with timeout let core fully start
      setTimeout(()=>{
        this._worker = new Threads.Worker(process.argv[1],{workerData:data})
          .on('message',msg=>{
            this.onThreadMessage.next(msg)
          })
          .on('error',err=>{
            log.err('error happened!')
            console.log(err)
            this.onThreadError.next(err)
          })
          .on('exit',res=>{
            log.warn('thread exited',res)
            this.M('events').pushEvent({
              type: 'interface:offline',
              data: {
                iface: this.f('code')
              }
            })
            this._worker = null
            this.onThreadStop.next(res)
            this.onThreadMessage.complete()
            this.onThreadError.complete()
            this.onThreadStop.complete()
            //this.onThread
          })
          .on('online',res=>{
            log.success('thread started')
            this.onThreadStart.next(res)
            this.onThreadStart.complete()
            this.M('events').pushEvent({
              type: 'interface:online',
              data: {
                iface: this.f('code')
              }
            })
            this.obsEnd(obs,res)
          })
        this.M('events').addWorker(this._worker)
      },1000)
    })
  }

  stopThread()
  {
    return Rx.Observable.create(obs=>{
      let nx = Rx.of(true)
      if (this._worker)
        nx = Rx.from(this._worker.terminate()).pipe(RxO.map(_=>true))
      nx.subscribe(res=>this.obsEnd(obs,res),err=>this.obsErr(obs,err))
    })
    /*
    return Rx.iif(_=>this._worker&&typeof this._worker.terminate=='function',
      Rx.of(this._wo
      Rx.throwError({code:'THREAD_NOT_STARTED'})
    )
      /**/
  }

  postThreadMessage(msg)
  {
    return Rx.Observable.create(obs=>{
      if (!this._worker)
        return this.obsErr(obs,{code:'WORKER_NOT_EXISTS'})
      this._worker.postMessage(msg)
      this.obsEnd(obs)
    })
  }

  //used from child thread
  postParentMessage(msg)
  {
    return Rx.Observable.create(obs=>{
      if (!Threads.parentPort)
        return this.obsErr(obs,{code:'WORKER_NOT_EXISTS'})
      Threads.parentPort.postMessage(msg)
      this.obsEnd(obs)
    })
  }




  /*
  //contacts
  ifContacts()
  {
    return Rx.throwError({code:'ABSTRACT_IFACE'})
  }

  ifHistory(user)
  {
  }

  ifMessage()
  {
  }

  ifSend()
  {
    return Rx.throwError({code:'ABSTRACT_IFACE'})
  }
  /**/

}

module.exports = AbstractInterface
