/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


/**
@fileOverview Module manager class file
@author abrakadobr
*/
let fs = require('fs')
let DT = require('directory-tree')
let Rx = require('rxjs')
let RxO = require('rxjs/operators')

let log = require('../../log').m('ModuleManager')

const CoreModule = require('../core/abstract/module')

/**
Core Module Manager class - provides modules loading, installing and so on
@class ModuleManager
@extends CoreModule
*/
class ModuleManager extends CoreModule
{
  constructor(parent)
  {
    super(parent)
    this._core = parent

    this._mdb = null
    this._lastInfoTree = null

    this._modules = {
      module_manager: {
        info: null,
        module: this,
        installed: false,
        loaded: true,
        enabled: true
      }
    }
  }

  distUpgrade(itree)
  {
    return Rx.Observable.create(obs=>{
      let $load = []
      let $inst = []
      let $defs = []
      let $init = []
      let appCfg = null
      try {
        appCfg = JSON.parse(fs.readFileSync('./config/app.conf.json').toString())
      } catch(e) {
        appCfg = null
      }
      itree.coreOrder.forEach(mcode=>{
        $load.push(this.loadModule(itree.core[mcode]))
        let cfg = null
        if (mcode=='db')
        {
          try {
            cfg = JSON.parse(fs.readFileSync('./config/db.conf.json').toString())
          } catch(e) {
            cfg = null
          }
          if (cfg&&cfg.redis)
            cfg.httpredis = {
              host: cfg.redis.host,
              port: cfg.redis.port,
              db: cfg.redis.db+1
            }
        }
        //log confs
        if (mcode=='log'&&appCfg&&appCfg.log)
          cfg = appCfg.log
        //file confs
        if (mcode=='file'&&appCfg&&appCfg.files)
          cfg = appCfg.files
        if (mcode=='dashboard'&&appCfg)
        {
          cfg = {
            host: '0.0.0.0',
            port: appCfg.PORT,
            sessionSecret: appCfg.sessionSecret
          }
        }

        let inst = null

        if (mcode.startsWith('http')||mcode=='menu'||mcode=='content')
        {
          inst = this.installModule(mcode)
          $defs.push(this.installModuleDefaults(mcode))
        } else {
          inst = this.installModule(mcode,{
            onlyMDB:true,
            enabled: true,
            config:cfg
          })
          $init.push(this.initModule(mcode))
        }
        if (mcode=='db')
          $inst.push(inst.pipe(
            RxO.mergeMap(x=>this.configureModule(mcode,x)),
            RxO.mergeMap(x=>this.initModule(mcode))
          ))
        else
          $inst.push(inst.pipe(RxO.mergeMap(x=>this.configureModule(mcode,x))))
      })
      Rx.concat(...$load,...$inst,...$init,...$defs).pipe(RxO.last()).subscribe(_=>{
        log.success('Dist Upgrade success')
        log.info('...')
        log.info('You can remove old config/*.conf.json files and restart normally.')
        log.warn('config/modules.json is new config file and it is required!')
        this.obsErr(obs,{code:'NORMAL_RESTART_REQUIRED'})
      },err=>this.obsErr(obs,err))
    })
  }

  installAndStart(mcode,conf=null)
  {
    return Rx.of(mcode).pipe(
      RxO.mergeMap(_=>this.buildInfoTree()),
      RxO.mergeMap(itree=>{
        if (!itree.order.includes(mcode))
          return Rx.throwError({code:'MODULE_NOT_FOUND',module:mcode})
        let minfo = null
        if (itree.core[mcode])
          minfo = itree.core[mcode]
        if (itree.user[mcode])
          minfo = itree.user[mcode]
        if (!minfo)
          return Rx.throwError({code:'MODULE_INFO_NOT_FOUND',module:mcode})
        return this.loadModule(minfo)
      }),
      RxO.mergeMap(_=>this.installModule(mcode,{config:conf})),
      RxO.mergeMap(cfg=>this.configureModule(mcode,cfg)),
      RxO.mergeMap(_=>this.initModule(mcode)),
      RxO.mergeMap(_=>this.installModuleDefaults(mcode)),
      RxO.mergeMap(_=>this.startModule(mcode))
    )
  }

  loadAndStart(tree)
  {
    return Rx.Observable.create(obs=>{
      let $load = []
      let $configure = []
      let $init = []
      let $start = []
      let $updates = []
      //console.log(tree.coreUpdates)
      tree.coreOrder.forEach(mcode=>{
        let cfg = null
        if (this._mdb&&this._mdb.core[mcode]&&this._mdb.core[mcode].config)
          cfg = this._mdb.core[mcode].config
        if (mcode=='log')
        {
          $load.push(this.loadModule(tree.core[mcode]).pipe(
            RxO.mergeMap(_=>this.configureModule(mcode,cfg))
          ))
        } else {
          $load.push(this.loadModule(tree.core[mcode]))
          $configure.push(this.configureModule(mcode,cfg))
        }
        $init.push(this.initModule(mcode))
        $start.push(this.startModule(mcode))
        let modUpd = tree.coreUpdates.reduce((acc,upd)=>upd.mod==mcode?upd:acc,null)
        if (modUpd)
          $updates.push(this.updateModule(modUpd))
      })
      tree.userOrder.forEach(mcode=>{
        $load.push(this.loadModule(tree.user[mcode]))
        let cfg = null
        if (this._mdb&&this._mdb.user[mcode]&&this._mdb.user[mcode].config)
          cfg = this._mdb.user[mcode].config
        $configure.push(this.configureModule(mcode))
        $init.push(this.initModule(mcode))
        $start.push(this.startModule(mcode))
        let modUpd = tree.userUpdates.reduce((acc,upd)=>upd.mod==mcode?upd:acc,null)
        if (modUpd)
          $updates.push(this.updateModule(modUpd))
      })
      Rx.concat(
        Rx.concat(...$load).pipe(RxO.last()),
        Rx.concat(...$configure).pipe(RxO.last()),
        Rx.concat(...$init).pipe(
          RxO.last(),
          RxO.mergeMap(_=>{
            if (!$updates.length)
              return Rx.of(null)
            return Rx.concat(...$updates).pipe(RxO.last())
          })
        ),
        Rx.concat(...$start).pipe(RxO.last())
      ).pipe(RxO.last()).subscribe(()=>{
        this.obsEnd(obs)
      },err=>this.obsErr(obs,err))
    })
  }

  installCore(itree)
  {
    return Rx.Observable.create(obs=>{
      let $load = []
      let $inst = []
      let $defs = []
      let $t = itree.coreOrder.map(x=>{
        $load.push(this.loadModule(itree.core[x]))
        $inst.push(this.installModule(x).pipe(
          RxO.mergeMap(xx=>this.configureModule(x,xx))
        ))
        $defs.push(this.installModuleDefaults(x))
      })
      Rx.concat(...$load,...$inst,...$defs).pipe(RxO.last()).subscribe(()=>{
        this.obsErr(obs,{code:'NORMAL_RESTART_REQUIRED',data:'install complete. restart normally'})
      },err=>this.obsErr(obs,err))
    })
  }

  buildWorkingTree(itree=null)
  {
    return Rx.Observable.create(obs=>{
      log.info('building working tree')
      let mdb = this._readMDB()

      let coreMods = itree.coreMods.filter(x=>{
        if (!mdb.core[x.fullCode])
          return false
        if (mdb.core[x.fullCode].installed&&mdb.core[x.fullCode].enabled)
          return true
        return false
      })
      let userMods = itree.userMods.filter(x=>{
        if (!mdb.user[x.fullCode])
          return false
        if (mdb.user[x.fullCode].installed&&mdb.user[x.fullCode].enabled)
          return true
        return false
      })

      let coreDeps = this._treeLevel(coreMods)
      let coreOrder = this._orderDepsLevel(coreDeps)
      let coreFail = coreMods.filter(mod=>!coreOrder.includes(mod.fullCode))

      let userDeps = this._treeLevel(userMods,'',[...coreOrder])
      let userOrder = this._orderDepsLevel(userDeps)
      let userFail = userMods.filter(mod=>!userOrder.includes(mod.fullCode))

      let coreUpdates = []
      let coreIndex = coreMods.reduce((acc,x)=>{
        let modcv = x.info.version
        let modiv = mdb.core[x.fullCode].version
        if (modcv!=modiv)
          coreUpdates.push({
            mod: x.fullCode,
            from: modiv,
            to: modcv
          })
        acc[x.fullCode] = x
        return acc
      },{})
      let userUpdates = []
      let userIndex = userMods.reduce((acc,x)=>{
        let modcv = x.info.version
        let modiv = mdb.user[x.fullCode].version
        if (modcv!=modiv)
          userUpdates.push({
            mod: x.fullCode,
            from: modiv,
            to: modcv
          })
        acc[x.fullCode] = x
        return acc
      },{})

      let wtree = {
        core: coreIndex,
        user: userIndex,
        coreMods: coreMods,
        userMods: userMods,
        coreTree: coreDeps,
        userTree: userDeps,
        coreUpdates: coreUpdates,
        userUpdates: userUpdates,
        coreOrder: coreOrder,
        userOrder: userOrder,
        coreFail: coreFail,
        userFail: userFail,
        order: [...coreOrder,...userOrder],
        fail: [...coreFail,...userFail]
      }
      //console.log(['wtree',wtree.order])
      //log.info('modules working tree size',wtree.order.length)
      log.info('required modules information tree builded. <core/user>',[wtree.coreMods.length,wtree.userMods.length])
      if (wtree.coreFail.length)
        log.warn('unresolved core modules exists!',wtree.coreFail.map(x=>x.fullCode)) 
      if (wtree.userFail.length)
        log.warn('unresolved user modules exists!',wtree.userFail.map(x=>x.fullCode)) 
      this.obsEnd(obs,wtree)
    })
  }

  buildInfoTree()
  {
    return Rx.Observable.create(obs=>{
        
      log.info('building info tree')

      let allMods = []

      let coreDir = DT('./core/modules',{extensions:/\.json/})
      let coreMods = []
      coreDir.children.forEach(cld=>{
        let list = this._cld2tree(cld.name,'core')
        if (!Array.isArray(list))
        {
          log.warn('not a module!',cld.path)
          return
        }
        list.forEach(x=>{
          coreMods.push(x)
          allMods.push(x)
        })
      })

      let userDir = DT('./modules',{extensions:/\.json/})
      let userMods = []
      userDir.children.forEach(cld=>{
        let list = this._cld2tree(cld.name,'user')
        if (!Array.isArray(list))
        {
          log.warn('not a module!',cld.path)
          return
        }
        list.forEach(x=>{
          userMods.push(x)
          allMods.push(x)
        })
      })

      /**
      let themesDir = DT('./themes',{extensions:/\.json/})
      let themesMods = []
      themesDir.children.forEach(cld=>{
        let list = this._cld2tree(cld.name,'theme')
        if (!Array.isArray(list))
        {
          log.warn('not a module!',cld.path)
          return
        }
        list.forEach(x=>{
          themesMods.push(x)
          allMods.push(x)
        })
      })
      /**/

      let order = []
      let nondep = []
      let deps = {}
      let groups = {}

      let coreDeps = this._treeLevel(coreMods)
      let coreOrder = this._orderDepsLevel(coreDeps)
      let coreFail = coreMods.filter(mod=>!coreOrder.includes(mod.fullCode))

      let userDeps = this._treeLevel(userMods,'',[...coreOrder])
      let userOrder = this._orderDepsLevel(userDeps)
      let userFail = userMods.filter(mod=>!userOrder.includes(mod.fullCode))

      //let themesDeps = this._treeLevel(themesMods,[...coreOrder,...userOrder])

      //console.log(JSON.stringify(this._logDepsLevel(coreDeps),null,2),coreOrder)
      //console.log(JSON.stringify(this._logDepsLevel(userDeps),null,2),userOrder)
      //console.log(JSON.stringify(this._logDepsLevel(themesDeps),null,2))
      if (coreFail.length)
      {
        log.err('core modules fail to load',coreFail.map(x=>x.fullCode))
      }
      let coreIndex = coreMods.reduce((acc,x)=>{
        acc[x.fullCode] = x
        return acc
      },{})
      let userIndex = userMods.reduce((acc,x)=>{
        acc[x.fullCode] = x
        return acc
      },{})

      let ret = {
        core: coreIndex,
        user: userIndex,
        coreMods: coreMods,
        userMods: userMods,
        coreTree: coreDeps,
        userTree: userDeps,
        coreOrder: coreOrder,
        userOrder: userOrder,
        coreFail: coreFail,
        userFail: userFail,
        order: [...coreOrder,...userOrder],
        fail: [...coreFail,...userFail]
      }
      log.info('all modules information tree builded. <core/user>',[ret.coreMods.length,ret.userMods.length])
      if (ret.coreFail.length)
        log.warn('unresolved core modules exists!',ret.coreFail.map(x=>x.fullCode)) 
      if (ret.userFail.length)
        log.warn('unresolved user modules exists!',ret.userFail.map(x=>x.fullCode)) 
      //console.log(['core modules info tree',ret.core.map(x=>x.fullCode),coreDeps])
      this._lastInfoTree = ret
      this.obsEnd(obs,ret)
    })
  }

  _orderDepsLevel(lvl)
  {
    let ret = []
    let next = []
    lvl.forEach(mod=>{
      let mods = []
      mods.push(mod.mod.fullCode)
      let nxc = []
      let nxn = []
      if (mod.chlds.length)
        nxc = this._orderDepsLevel(mod.chlds)
      if (mod.next.length)
        nxn = this._orderDepsLevel(mod.next)
      //if (mod.next.length)
        //ret[mod.mod.fullCode] = this._logDepsLevel(mod.next)
      if (nxc.length)
        mods = [...mods,...nxc]
      if (nxn.length)
        next = [...next,...nxn]
      ret = [...ret,...mods]
    })
    return [...ret,...next]
  }

  _logDepsLevel(lvl)
  {
    let ret = {}
    lvl.forEach(mod=>{
      ret[mod.mod.fullCode] = 'ok'
      let nxc = {}
      let nxn = {}
      if (mod.chlds.length)
        nxc = this._logDepsLevel(mod.chlds)
      if (mod.next.length)
        nxn = this._logDepsLevel(mod.next)
      if (Object.keys(nxc).length||Object.keys(nxn).length)
      {
        let obj = {}
        if (Object.keys(nxc).length)
          obj.childs = nxc
        if (Object.keys(nxn).length)
          obj.next = nxn
        ret[mod.mod.fullCode] = obj
      }
    })
    return ret
  }

  _treeLevel(mods,lvl='',loaded=[],childs=false)
  {
    let ret = []
    //console.log(mods)
    mods.forEach(mod=>{
      let ok = false
      if (mod.deps.includes(lvl))
        ok = true
      if (lvl=='')
      {
        if (!loaded.length)
        {
          if (!mod.deps.length)
            ok = true
        } else {
          if (mod.deps.length)
            mod.deps.forEach(dep=>{
              if (loaded.includes(dep))
                ok = true
            })
          else
            ok = true
        }
      } else {
        if (childs)
        {
          if (mod.parent!=lvl)
            ok = false
        } else {
          if (mod.parent==lvl)
            ok = false
        }
      }
      if (loaded.includes(mod.fullCode))
        ok = false
      if (ok&&mod.deps.length)
        mod.deps.forEach(mdep=>{
          if (!loaded.includes(mdep))
            ok = false
        })
      if (ok)
      {
        loaded.push(mod.fullCode)
        let cMods = this._treeLevel(mods,mod.fullCode,loaded,true)
        let nMods = this._treeLevel(mods,mod.fullCode,loaded,false)
        ret.push({
          mod: mod,
          chlds: cMods,
          next: nMods
        })
      }
    })
    return ret
  }

  _cld2tree(code,scope,prnt='')
  {
    let dirPath = 'modules/'
    if (scope=='core')
      dirPath = 'core/modules/'
    if (scope=='theme')
      dirPath = 'themes/'
    if (prnt&&prnt.length&&scope!='theme')
      dirPath += prnt+'/'
    dirPath += code+'/'
    let jsonPath = dirPath
    if (scope=='theme')
      jsonPath +='theme.json'
    else
      jsonPath += code+'.json'
    let info = null
    try {
      info = JSON.parse(fs.readFileSync(jsonPath).toString())
    } catch(e) {
      info = null
    }
    if (!info)
      return null
    if (!fs.existsSync(dirPath+'module.js'))
      return null
    if (!Array.isArray(info.deps))
      info.deps = []
    //if (scope=='theme'&&!info.deps.includes('http/theme'))
      //info.deps = ['http/theme',...info.deps]
    let ret = {
      code: code,
      fullCode: code,
      path: jsonPath,
      dir: dirPath,
      info: info,
      scope: scope,
      parent: prnt,
      childs: [],
      deps: info.deps
    }
    if (prnt&&prnt.length)
    {
      ret.fullCode = prnt+'/'+code
      if (!ret.deps.includes(prnt))
        ret.deps = [prnt,...info.deps]
    }
    if (scope=='theme')
    {
      ret.fullCode = 'theme/'+code
      if (info.base&&info.base.length)
        ret.deps.push('theme/'+info.base)
    }
    let retlist = []
    if (Array.isArray(info.childs)&&info.childs.length)
    {
      ret.childs = info.childs.map(chld=>{
        let childslist = this._cld2tree(chld,scope,ret.fullCode)
        //console.log(['chldsList',chld,scope,code,childslist])
        if (Array.isArray(childslist))
          childslist.forEach(x=>{
            retlist.push(x)
          })
      })
    }
    return [ret,...retlist]
  }

  moduleExists(code) { return typeof this._modules[code] != 'undefined' }
  moduleLoaded(code) {
    return this.moduleExists(code)&&this._modules[code].loaded
  }
  moduleInstalled(code) {
    return this.moduleExists(code)&&this._modules[code].installed
  }
  moduleEnabled(code) {
    return this.moduleExists(code)&&this._modules[code].enabled
  }
  moduleObject(code) {
    if (!this.moduleExists(code)) return null
    return this._modules[code].module
  }

  loadModule(info,installed=false)
  {
    return Rx.Observable.create(obs=>{
      if (info.fullCode=='module_manager')
      {
        this._modules.module_manager.info = info
        return this.obsEnd(obs,info)
      }
      log.log('loading module',info.fullCode)
      let MCls = null
      try {
        MCls = require('../../../'+info.dir+'module')
      } catch(e) {
        return this.obsErr(obs,{code:'MODULE_CLASS_REQUIRE_ERROR',err:e,module:info.fullCode})
      }
      if (!CoreModule.isPrototypeOf(MCls))
        return this.obsErr(obs,{code:'MODULE_CLASS_PROTOTYPE_ERROR',module:info.fullCode})
      log.log('loading module cls',info.fullCode)
      let md = new MCls(this._core)
      md.moduleLoad().subscribe(_=>{
        log.log('module loaded',[info.fullCode,info.info.version])
        this._modules[info.fullCode] = {
          info: info,
          module: md,
          loaded: true,
          installed: installed,
          enabled: true
        }
        this.obsEnd(obs)
      },err=>{
        this.obsErr(obs,err)
      })
    })
  }

  installModule(mcode,opts={})
  {
    return Rx.Observable.create(obs=>{
      //console.log(['installModule',mcode])
      if (!this.moduleLoaded(mcode))
      {
        console.log([mcode,this._modules[mcode]])
        return this.obsErr(obs,{code:'MODULE_NOT_LOADED',module:mcode,fn:'installModule'})
      }
      let minfo = this._modules[mcode].info

      log.warn('installing module',mcode)
      let mdb = this._readMDB()
      mdb[minfo.scope][mcode] = {
        version: minfo.info.version,
        installed: false,
        enabled: false,
        config: null
      }
      if (opts&&opts.enabled)
        mdb[minfo.scope][mcode].enabled = true
      if (opts&&opts.onlyMDB)
      {
        mdb[minfo.scope][mcode].installed = true
        if (opts&&opts.config)
          mdb[minfo.scope][mcode].config = opts.config
        this._writeMDB(mdb)
        return this.obsEnd(obs,mdb[minfo.scope][mcode].config)
      }
      this._writeMDB(mdb)

      this.moduleObject(mcode).moduleInstall().subscribe(modCfg=>{
        log.log('module install routine complete!',mcode)
        let mdb2 = this._readMDB()
        mdb2[minfo.scope][mcode].installed = true
        mdb2[minfo.scope][mcode].enabled = true
        mdb2[minfo.scope][mcode].config = modCfg
        if (opts&&opts.config)
          mdb2[minfo.scope][mcode].config = opts.config
        this._writeMDB(mdb2)
        this._modules[mcode].installed = true
        this.obsEnd(obs,mdb2[minfo.scope][mcode].config)
      },err=>{
        console.log(err)
        this.obsErr(obs,err)
      })
    })
  }

  installModuleDefaults(mcode)
  {
    return Rx.Observable.create(obs=>{
      //console.log(['installModuleDefaults',mcode,this._modules[mcode]])
      if (!this.moduleInstalled(mcode))
        return this.obsErr(obs,{code:'MODULE_NOT_INSTALLED',module:mcode,fn:'installModuleDefaults'})
      this.moduleObject(mcode).moduleDefaults().subscribe(_=>this.obsEnd(obs),err=>this.obsErr(obs,err))
    })
  }


  /**/

  configureModule(code,conf=null)
  {
    return Rx.Observable.create(obs=>{
      if (!this.moduleLoaded(code))
        return this.obsErr(obs,{code:'CONFIGURE_MODULE_NOT_LOADED',module:code})
      log.log('configuring',code)
      this.moduleObject(code).moduleConfigure(conf).subscribe(()=>{
        log.log('module configured',code)
        this.obsEnd(obs)
      },err=>this.obsErr(obs,err))
    })
  }

  updateModule(upd)
  {
    return Rx.Observable.create(obs=>{
      let code = upd.mod
      if (!this.moduleLoaded(code))
        return this.obsErr(obs,{code:'UPDATE_MODULE_NOT_LOADED',module:code})
      log.warn('updating',upd)
      this.moduleObject(code).moduleUpdate(upd.from,upd.to).subscribe(ver=>{
        log.log('module updated',[code,ver])
        let mdb = this._readMDB()
        let done = false
        Object.keys(mdb.core).forEach(mcode=>{
          if (mcode!=code)
            return
          mdb.core[code].version = ver
          this._writeMDB(mdb)
          done = true
        })
        if (done)
          return this.obsEnd(obs)
        Object.keys(mdb.user).forEach(mcode=>{
          if (mcode!=code)
            return
          mdb.core[code].version = ver
          this._writeMDB(mdb)
          done = true
        })
        return this.obsEnd(obs)
      },err=>this.obsErr(obs,err))
    })
  }

  initModule(code)
  {
    return Rx.Observable.create(obs=>{
      if (!this.moduleLoaded(code))
        return this.obsErr(obs,{code:'INIT_MODULE_NOT_LOADED',module:code})
      log.log('initializing',code)
      this.moduleObject(code).moduleInit().subscribe(()=>{
        log.log('module initialized',code)
        this.obsEnd(obs)
      },err=>this.obsErr(obs,err))
    })
  }

  startModule(code)
  {
    return Rx.Observable.create(obs=>{
      if (!this.moduleLoaded(code))
        return this.obsErr(obs,{code:'START_MODULE_NOT_LOADED',module:code})
      log.log('starting',code)
      this.moduleObject(code).moduleStart().subscribe(()=>{
        log.log('module started',code)
        this.obsEnd(obs)
      },err=>this.obsErr(obs,err))
    })
  }

  /**/

  _readMDB()
  {
    if (this._mdb)
      return this._mdb
    let db = null
    try {
      db = JSON.parse(fs.readFileSync('./config/modules.json').toString())
    } catch(e) {
      db = null
    }
    if (!db)
      db = {
        core: {},
        user: {},
        themes: {}
      }
    this._mdb = db
    return db
  }

  _writeMDB(db)
  {
    this._mdb = db
    fs.writeFileSync('./config/modules.json',JSON.stringify(db,null,2))
  }

  /**/

  hookCollect(hook,...data)
  {
    return Rx.Observable.create(obs=>{
      let $t = []
      for(let code in this._modules)
      {
        if (!this._modules[code].enabled) continue
        let mdl = this.moduleObject(code)
        //console.log([this._modules[code]])
        if (typeof mdl[hook] == 'function')
          $t.push(mdl[hook](...data))
      }
      let acc = []
      Rx.concat(...$t).subscribe(val=>{
        acc.push(val)
      },err=>{
        log.err('hookCollect failed',hook)
        console.log(err)
        obs.next(acc)
        obs.complete()
      },()=>{
        obs.next(acc)
        obs.complete()
      })
    })
  }

  hookMap(hook,data)
  {
    return Rx.Observable.create(obs=>{
      let $chain = null
      for(let code in this._modules)
      {
        if (!this._modules[code].enabled) continue
        let mdl = this.moduleObject(code)
        if (typeof mdl[hook] == 'function')
        {
          if (!$chain)
            $chain = mdl[hook](data)
          else
            $chain = $chain.pipe(RxO.mergeMap(x=>{
              //log.test('chaining hookMap',hook)
              return mdl[hook](x)
            }))
        }
      }
      if (!$chain)
        return this.obsEnd(obs,data)
      $chain.subscribe(fin=>{
        this.obsEnd(obs,fin)
      },err=>{
        log.err('hookMap failed',hook)
        console.log(err)
        this.obsErr(obs,{code:'HOOK_MAP_ERROR',error:err})
      })
    })
  }


  dashboardRoutes(app)
  {
    return Rx.Observable.create(obs=>{

      let rtList = {
        url: '/api/mm/list',
        access: 'global.root',
        method: 'get',
        fn: (ctx)=>{
          return Rx.of(ctx).pipe(
            RxO.mergeMap(_=>this.buildInfoTree()),
            RxO.map(itree=>{
              let mdb = this._readMDB()
              return {
                itree: itree,
                mdb: mdb
              }
            })
          )
        }
      }

      let rtConf = {
        url: '/api/mm/conf',
        access: 'global.root',
        method: 'post',
        fn: (ctx)=>{
          return Rx.of(ctx).pipe(
            /*
            RxO.mergeMap(_=>this.buildInfoTree()),
            RxO.map(itree=>{
              let mdb = this._readMDB()
              return {
                itree: itree,
                mdb: mdb
              }
            }),
            /**/
            RxO.mergeMap(data=>{
              let b = ctx.f('body')
              let mcode = b.mcode
              let mmdb = b.mdb
              let mdb = this._readMDB()
              let mdbi = null
              if (mdb.core[mcode])
                mdbi = mdb.core[mcode]
              if (!mdbi&&mdb.user[mcode])
                mdbi = mdb.user[mcode]
              
              //console.log(mcode,mmdb,mdbi)

              if (!mdbi||!mdbi.installed) //install
                return this.installAndStart(mcode,mmdb.config).pipe(RxO.map(x=>{
                  return {success:true}
                })).pipe(RxO.catchError(x=>{
                  return {success:false,error:x}
                }))

              return this.configureModule(mcode,mmdb.config).pipe(RxO.map(x=>{
                if (mdb.core[mcode])
                  mdb.core[mcode].config = mmdb.config
                if (mdb.user[mcode])
                  mdb.user[mcode].config = mmdb.config
                this._writeMDB(mdb)
                return {success:true}
              })).pipe(RxO.catchError(x=>{
                return {success:false, error: x}
              }))

            })
          )
        }
      }

      this.obsEnd(obs,[
        rtList,
        rtConf
      ])
    })
  }

}

module.exports = ModuleManager
