/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxOps = require('rxjs/operators')
let sanitizeHtml = require('sanitize-html')

let log = require('../../log').m('TelegramBot')
let g18n = require('../../g18n')

const CoreObject = require('../core/abstract/object')

const Telegraf = require('telegraf')
const Extra = require('telegraf/extra')
const Markup = require('telegraf/markup')
const tgSession = require('telegraf/session')

const BotQueue = require('./botQueue')

class Bot extends CoreObject
{
  constructor(parent)
  {
    super(parent)
    this._bot = null
    this._router = null
    this._token = null
    this._username = ''
    this._connected = false

    this._profile = null

    this._config = {
      username: '',
      token: '',
      isLongPoll: true,
      webhookHost: '',
      webhookPort: 0
    }

    this.signals.connected = new Rx.Subject()
  }

  setConnected(val)
  {
    this._connected = val
    this.signals.connected.next(this._connected)
  }

  configure(cfg)
  {
    this._config.username = cfg.username
    this._config.token = cfg.token
    this._config.isLongPoll = true
    if (!cfg.isLongPoll&&cfg.webhookHost!=''&&cfg.webhookPort>0)
    {
      this._config.isLongPoll = false
      this._config.webhookHost = cfg.webhookHost
      this._config.webhookPort = cfg.webhookPort
    }
    return true
  }

  init()
  {
    if (!this._config.token||!this._config.username)
    {
      log.err('not configured')
      return
    }
    this._bot = new Telegraf(this._config.token)
    this._queue = new BotQueue(this)
    this._bot.use(tgSession({
      getSessionKey: (ctx) => {
        if (ctx.from && ctx.chat) {
          return `${ctx.from.id}:${ctx.chat.id}`
        } else if (ctx.from && ctx.inlineQuery) {
          return `${ctx.from.id}:${ctx.from.id}`
        }
        return null
      }
    }))
    this._queue.next.subscribe($msg=>{
      //now REAL send it
      this._sendMsg($msg)
    })

  }

  reconfigure(token)
  {
    return Rx.Observable.create(obs=>{
      log.warn('tesing new token',token)
      this.stop()
      this._bot.token = token
      this.loadProfile().subscribe(()=>{
        log.success('reconfigured successfully')
        this.start()
        this.obsEnd(obs,true)
      },err=>{
        log.err('reconfiguring failed,return ti previous token')
        this.stop()
        this._bot.token = this.M('config').conf('telegram.token')
        this.loadProfile().subscribe(()=>{
          log.info('previous profile loaded')
          this.start()
          this.obsEnd(obs,false)
        },err=>{
          log.err('return to previous token failed!!!')
          this.obsEnd(obs,false)
        })
      })
    })
  }

  start()
  {
    if (this._connected) return
    //this._bot.startPolling(3,5,['message','callback_query','inline_query','channel_post'],()=>{
    //this._bot.startPolling(3,5,()=>{
      //this.setConnected(false)
    //})
    let cfg = {}
    if (this._config.isWebhook)
      cfg.webhook = {
        domain: this._config.webhookDomain,
        port: this._config.webhookPort,
        cb: _=>{
          this.setConnected(false)
        }
      }
    else
      cfg.polling = {
        timeout: 10,
        limit: 50,
        stopCallback: _=>{
          this.setConnected(false)
        }
      }
    this._bot.launch(cfg)
    this.setConnected(true)
    this._queue.run(true)
  }

  stop()
  {
    if (!this._connected) return
    this._queue.run(false)
    this._bot.stop()
    this.setConnected(false)
  }

  use(mw)
  {
    this._bot.use(mw)
  }

  on(ev,fn)
  {
    this._bot.on(ev,fn)
  }

  profile()
  {
    return this._profile
  }

  loadProfile()
  {
    return Rx.Observable.create(obs=>{
      Rx.from(this._bot.telegram.getMe()).pipe(
        RxOps.mergeMap(info=>{
          return this.M('telegram').tgUser(info)
        })
      ).subscribe(self=>{
        this._profile = self
        if (!self._new)
          return this.obsEnd(obs)
        this._profile.save().subscribe(p=>{
          this.obsEnd(obs)
        },err=>{
          log.err('saving self profile on loading failed')
          console.log(err)
          this.obsErr(obs,err)
        })
      },err=>{
        log.err('loading self profile failed')
        console.log(err)
        this.obsErr(obs,err)
      })
    })
  }

  deleteMessages(list)
  {
    return Rx.Observable.create(obs=>{
      if (!Array.isArray(list)||!list.length)
        return this.obsEnd(obs)
      let $t = list.map(msg=>{
        //console.log(msg.toObject())
        return Rx.of(this._bot.telegram.deleteMessage(msg.fp('chat').f('chatId'),msg.f('messageId')))
      })
      Rx.concat(...$t).subscribe(_=>{},err=>this.obsErr(obs,err),_=>this.obsEnd(obs))
    })
  }

  kickUser(user,chat)
  {
    return Rx.Observable.create(obs=>{
      log.warn('user kick!',[user.titles(),user.f('tgid'),chat.f('title'),chat.f('chatId')])
      this._bot.telegram.kickChatMember(chat.f('chatId'),user.f('tgid'))
      this.obsEnd(obs)
    })
  }

  inlineReply(iqid,rep,extra)
  {
    //return Rx.Observable.create(obs=>{
      return Rx.from(this._bot.telegram.answerInlineQuery(iqid,rep,extra))
      //this.obsEnd(obs)
    //})
  }

  sendMessage($msg)
  {
    if (!$msg||!$msg.sendable())
      return
    //console.log(['botSendMEssage',$msg.fp('chat').f('title'),$msg.f('text')])
    //we don't SEND message here, just push it 
    //to sending queue
    this._queue.push($msg)
  }

  _sendMsg($msg)
  {
    if (!$msg||!$msg.sendable())
      return

    $msg.compile().subscribe(msg=>{
      if (!msg)
        return
      let send = {
        chatId: $msg.fp('chat').f('chatId'),
        msg: msg
      }
      if ($msg.fp('chat').f('type')=='channel'&&send.msg.kbdType=='none')
        send.msg.kbdType = 'parent'
      //console.log(['send',send])
      this.send2chat(send).subscribe(res=>{
        //console.log(['message send res',res])
        $msg.finilize(res).subscribe(m=>{
          //log.success('msg finilized')
          $msg.sendDone({success:true})
        },err=>{
          log.err('message finilization failed')
          //console.log(err)
          $msg.sendDone({success:true})
        })
      },err=>{
        log.err('send2chat message failed!')
        if (err&&err.description&&(err.description=='Forbidden: bot was blocked by the user'||err.description=='Forbidden: user is deactivated'))
        {
          let $chat = $msg.fp('chat')
          $chat.f('open',false)
          $chat.save().pipe(
            RxOps.mergeMap(_=>this.C('user').findNode({tgid:$chat.f('chatId')})),
            RxOps.mergeMap($u=>{
              if (!$u) return Rx.of(null)
              $u.field('privateChat').clear()
              if (err.description=='Forbidden: user is deactivated')
              {
                $u.f('lastName','[INACTIVE]')
                log.warn('User not active any more =(',$u.titles())
              } else {
                $u.f('lastName','[BANED]')
                log.warn('User baned this bot!',$u.titles())
              }
              return $u.save()
            }),
            RxOps.catchError(err=>{
              log.err('error happens on updating baned user!')
              console.log(err)
              return Rx.of(null)
            })
          ).subscribe(_=>{
            $msg.sendDone({success:false,error:true,err:err})
          })
        } else {
          console.log(err)
          $msg.sendDone({success:false,error:true,err:err})
        }
      })
    },err=>{
      log.err('message not compiled!')
      console.log(err)
      $msg.sendDone({success:false,error:true})
    })
  }

  //======= helpers
  sanitizeTgHtml(html)
  {
    //clear any unreplaced token
    //let str = html.replace(new RegExp('{[^}]+}','g'),'')
    let str = html
    //prepare sinonyms
    str = str.replace(/<strong>/g,'<b>')
    str = str.replace(/<\/strong>/g,'</b>')
    str = str.replace(/<em>/g,'<i>')
    str = str.replace(/<\/em>/g,'</i>')
    str = str.replace(/<p><br><\/p>/g,'<p></p>')
    str = str.replace(/<\/p><p>/g,'\n')
    log.log('Sanitized',str)
    //sanitize
    str = sanitizeHtml(str, {
      allowedTags: [ 'b', 'i', 'a', 'code', 'br' ],
      allowedAttributes: {
        'a': [ 'href' ]
      },
      allowedSchemes: [ 'http', 'https', 'tg', 'mailto' ],
    })
    //replace new lines
    str = str.replace(/<br[^>]*>/g,'\n')
    str = str.replace(/<br\/>/g,'\n')
    str = str.replace(/&nbsp;/g,' ')
    return str
  }

  sanitizeAllHtml(html)
  {
    //let ret = html.replace(new RegExp('{[^}]+}','g'),'')
    let str = html
    ret = ret.replace(/&nbsp;/g,' ')
    ret = ret.replace(/<[^>]+>/g,'')
    return ret
  }


  //===== message
  message2ChatHtml(chatId,msg,clearKeyboard=true,updId=null)
  {
    let extra = Extra.markup()
    if (clearKeyboard)
      extra = Markup.removeKeyboard().extra()
    extra.parse_mode = 'HTML'
    msg = this.sanitizeTgHtml(msg)
    //console.log(['bot.js','sending text message',msg.length,msg])
    if (msg&&msg.length)
      msg = msg.trim()
    if (!msg||!msg.length)
      return Rx.throwError({code:'TG_MESSAGE_TEXT_EMPTY'})
    if (!updId)
      return Rx.from(this._bot.telegram.sendMessage(chatId,msg,extra))
    return Rx.from(this._bot.telegram.editMessageText(chatId,updId,undefined,msg,extra))
  }

  message2ChatKeyboardHtml(chatId,msg,kbd,updId=null)
  {
    let extra = Markup.keyboard(kbd).oneTime().resize().extra()
    extra.parse_mode = 'HTML'
    msg = this.sanitizeTgHtml(msg)
    if (msg&&msg.length)
      msg = msg.trim()
    if (!msg||!msg.length)
      return Rx.throwError({code:'TG_MESSAGE_TEXT_EMPTY'})
    if (!updId)
      return Rx.from(this._bot.telegram.sendMessage(chatId,msg,extra))
    return Rx.from(this._bot.telegram.editMessageText(chatId,updId,undefined,msg,extra))
  }

  message2ChatInlineKeyboardHtml(chatId,msg,kbd,updId=null)
  {
    let extra = Markup.inlineKeyboard(kbd).extra()
    extra.parse_mode = 'HTML'
    msg = this.sanitizeTgHtml(msg)
    if (msg&&msg.length)
      msg = msg.trim()
    if (!msg||!msg.length)
      return Rx.throwError({code:'TG_MESSAGE_TEXT_EMPTY'})
    if (!updId)
      return Rx.from(this._bot.telegram.sendMessage(chatId,msg,extra))
    return Rx.from(this._bot.telegram.editMessageText(chatId,updId,undefined,msg,extra))
  }

  //===== image
  image2Chat(chatId, image,clearKeyboard=true)
  {
    let extra = Markup.removeKeyboard().extra()
    if (!clearKeyboard)
      extra = Markup.resize().extra()
    let img = { source: image.source }
    if (image.caption)
    {
      extra.caption = this.sanitizeTgHtml(image.caption)
      extra.parse_mode='HTML'
    }
    return Rx.from(this._bot.telegram.sendPhoto(chatId,img,extra))
  }

  image2ChatKeyboard(chatId,image,kbd)
  {
    let extra = Markup.keyboard(kbd).oneTime().resize().extra()
    let img = { source: image.source }
    if (image.caption)
    {
      extra.caption = this.sanitizeTgHtml(image.caption)
      extra.parse_mode='HTML'
    }
    return Rx.from(this._bot.telegram.sendPhoto(chatId,img,extra))
  }

  image2ChatInlineKeyboard(chatId,image,kbd)
  {
    let extra = Markup.inlineKeyboard(kbd).extra()
    let img = { source: image.source }
    if (image.caption)
    {
      extra.caption = this.sanitizeTgHtml(image.caption)
      extra.parse_mode='HTML'
    }
    return Rx.from(this._bot.telegram.sendPhoto(chatId,img,extra))
  }

  //===== gallery
  gallery2Chat(chatId, media,clearKeyboard=true)
  {
    let extra = Markup.removeKeyboard().extra()
    if (!clearKeyboard)
      extra = Markup.resize().extra()
    return Rx.from(this._bot.telegram.sendMediaGroup(chatId,media,extra))
  }

  gallery2ChatKeyboard(chatId,media,kbd)
  {
    let extra = Markup.keyboard(kbd).oneTime().resize().extra()
    return Rx.from(this._bot.telegram.sendMediaGroup(chatId,media,extra))
  }

  gallery2ChatInlineKeyboard(chatId,media,kbd)
  {
    let extra = Markup.inlineKeyboard(kbd).extra()
    return Rx.from(this._bot.telegram.sendMediaGroup(chatId,media,extra))
  }

  //===== file
  file2Chat(chatId, file,clearKeyboard=true)
  {
    let extra = Markup.removeKeyboard().extra()
    if (!clearKeyboard)
      extra = Markup.resize().extra()
    return Rx.from(this._bot.telegram.sendDocument(chatId,file,extra))
  }

  file2ChatKeyboard(chatId,file,kbd)
  {
    let extra = Markup.keyboard(kbd).oneTime().resize().extra()
    return Rx.from(this._bot.telegram.sendDocument(chatId,file,extra))
  }

  file2ChatInlineKeyboard(chatId,file,kbd)
  {
    let extra = Markup.inlineKeyboard(kbd).extra()
    return Rx.from(this._bot.telegram.sendDocument(chatId,file,extra))
  }

  send2chat(compiled)
  {
    let msg = compiled.msg
    let chatId = compiled.chatId
    if (msg.type == 'text2u')
    {
      if (!chatId&&msg.user&&msg.user.tgid)
        chatId=msg.user.tgid
      return this.message2ChatHtml(chatId,msg.text)
    }

    if (msg.type == 'text')
    {
      //console.log(['sending',msg])
      if (msg.kbdType == 'none')
        return this.message2ChatHtml(chatId,msg.text,true,msg.upd)
      if (msg.kbdType == 'parent')
        return this.message2ChatHtml(chatId,msg.text,false,msg.upd)
      if (msg.kbdType == 'keyboard')
        return this.message2ChatKeyboardHtml(chatId,msg.text,msg.kbd,msg.upd)
      if (msg.kbdType == 'inline')
        return this.message2ChatInlineKeyboardHtml(chatId,msg.text,msg.kbd,msg.upd)
    }
    if (msg.type == 'file')
    {
      if (msg.kbdType == 'none')
        return this.file2Chat(chatId,msg.file)
      if (msg.kbdType == 'parent')
        return this.file2Chat(chatId,msg.file,false)
      if (msg.kbdType == 'keyboard')
        return this.file2ChatKeyboard(chatId,msg.file,msg.kbd)
      if (msg.kbdType == 'inline')
        return this.file2ChatInlineKeyboard(chatId,msg.file,msg.kbd)
    }
    if (msg.type == 'image')
    {
      if (msg.kbdType == 'none')
        return this.image2Chat(chatId,msg.image)
      if (msg.kbdType == 'parent')
        return this.image2Chat(chatId,msg.image,false)
      if (msg.kbdType == 'keyboard')
        return this.image2ChatKeyboard(chatId,msg.image,msg.kbd)
      if (msg.kbdType == 'inline')
        return this.image2ChatInlineKeyboard(chatId,msg.image,msg.kbd)
    }
    if (msg.type == 'gallery')
    {
      if (msg.kbdType == 'none')
        return this.gallery2Chat(chatId,msg.media)
      if (msg.kbdType == 'parent')
        return this.gallery2Chat(chatId,msg.media,false)
      if (msg.kbdType == 'keyboard')
        return this.gallery2ChatKeyboard(chatId,msg.media,msg.kbd)
      if (msg.kbdType == 'inline')
        return this.gallery2ChatInlineKeyboard(chatId,msg.media,msg.kbd)
    }
    return Rx.of(null)
  }

}

module.exports = Bot
