/*
  This file is part of TakatanJS software.
  @author abrakadobr
  @version 2.5.58
  @website https://js.takatan.tk
*/


let Rx = require('rxjs')
let RxO = require('rxjs/operators')
let Liquid = require('liquidjs')
let sanitizeHtml = require('sanitize-html')
const rxsafe = require('safe-regex')

let log = require('../../log').m('TemplateModule')
let g18n = require('../../g18n')
let mtz = require('moment-timezone')

const CoreModule = require('../core/abstract/module')

class TemplateModule extends CoreModule
{

  constructor(parent)
  {
    super(parent)
    this._engine = Liquid()
    this._engine.registerFilter('ago',(strVal,ffrom,lcl)=>{
      let mmt = mtz(strVal,ffrom)
      if (!mmt.isValid())
        return strVal
      if (lcl)
        mmt.locale(lcl)
      let ret = mmt.fromNow()
      return ret
    })
    this._engine.registerFilter('moment',(strVal,ffrom,fto,lcl)=>{
      let mmt = mtz(strVal,ffrom)
      if (!mmt.isValid()||!fto)
        return strVal
      if (lcl)
        mmt.locale(lcl)
      let ret = mmt.format(fto)
      return ret
    })
    this._engine.registerFilter('maps',(data,...fields)=>{
      if (!Array.isArray(data))
      {
        let ret = {}
        Object.keys(data).forEach(key=>{
          if (fields.includes(key))
            ret[key] = data[key]
        })
        return ret
      }
      let nx = []
      data.forEach(obj=>{
        let nxo = {}
        Object.keys(obj).forEach(key=>{
          if (fields.includes(key))
            nxo[key] = obj[key]
        })
        nx.push(nxo)
      })
      return nx
    })
    this._engine.registerFilter('imgpcut',(strVal)=>{
      let rx = /(<p>(<img[^>]*>)<\/p>)/gi 
      let m                
      let ret = strVal     
      while ((m = rx.exec(ret)) !== null) { 
        // This is necessary to avoid infinite loops with zero-width matches
        if (m.index === rx.lastIndex) { 
            rx.lastIndex++;
        }                  
        if (m&&m[1].length&&m[2].startsWith('<img'))
          ret = ret.replace(m[1],m[2])  
      }
      return ret           
    })

  }

  compile(template,ctx,as=null)
  {
    return Rx.Observable.create(obs=>{
      setTimeout(()=>{
        //console.log(ctx.view().virtual.asd)
        if (!template||!template.length)
          return this.obsEnd(obs,'')
        template = template.replace(new RegExp('&gt;','g'),'>')
        template = template.replace(new RegExp('&lt;','g'),'<')
        template = template.replace(/<p><br><\/p>/g,'<p></p>')
        template = template.replace(/<\/p><p>/g,'\n')
        template = template.replace(/<br[^>]*>/g,'\n')
        template = template.replace(/<br\/>/g,'\n')
        //console.log(template)
        let obj = ctx.view()
        if (as&&typeof as=='string')
        {
          obj = {}
          obj[as] = ctx.view(true)
        }
        //if (as)
          //console.log(template,obj)
        Rx.from(this._engine.parseAndRender(template,obj)).subscribe(txt=>{
          //console.log(txt)
          this.obsEnd(obs,txt)
        },err=>{
          log.err('compiling failed')
          console.log(err)
          this.obsEnd(obs,err)
        })
      },0)
    })
  }

  liquify(tpl,data)
  {
    return Rx.Observable.create(obs=>{
      //log.info('liquifying',tpl)
      //console.log(['liquify',tpl,data])
      if (typeof tpl == 'undefined')
        tpl = ''
      if (tpl=='')
        return this.obsEnd(obs,'')
      Rx.from(this._engine.parseAndRender(tpl,data)).subscribe(x=>{
        this.obsEnd(obs,x)
      },err=>{
        log.err('error templaing')
        console.log(err)
        this.obsEnd(obs,err)
      })
    })
  }

  str2rx(str,keys='')
  {
    let isRx=/^\/(.+)\/([a-z]*)/
    let m = isRx.exec(str)
    let ret = null
    if (m)
      ret = new RegExp(m[1],m[2]+keys)
    else
      ret = new RegExp(str,keys)
    if (!rxsafe(ret))
    {
      log.err('NOT SAFE REGEXP',[str,keys])
      ret = null
    }
    return ret
  }

  //======= helpers
  sanitizeTgHtml(html)
  {
    //clear any unreplaced token
    //let str = html.replace(new RegExp('{[^}]+}','g'),'')
    let str = html
    //prepare sinonyms
    str = str.replace(/<strong>/g,'<b>')
    str = str.replace(/<\/strong>/g,'</b>')
    str = str.replace(/<em>/g,'<i>')
    str = str.replace(/<\/em>/g,'</i>')
    str = str.replace(/<p><br><\/p>/g,'<p></p>')
    str = str.replace(/<\/p><p>/g,'\n')
    log.log('Sanitized',str)
    //sanitize
    str = sanitizeHtml(str, {
      allowedTags: [ 'b', 'i', 'a', 'code', 'br' ],
      allowedAttributes: {
        'a': [ 'href' ]
      },
      allowedSchemes: [ 'http', 'https', 'tg', 'mailto' ],
    })
    //replace new lines
    str = str.replace(/<br[^>]*>/g,'\n')
    str = str.replace(/<br\/>/g,'\n')
    str = str.replace(/&nbsp;/g,' ')
    return str
  }

  sanitizeAllHtml(html)
  {
    //let ret = html.replace(new RegExp('{[^}]+}','g'),'')
    let ret = html
    if (!ret||!ret.length)
      ret = ''
    ret = ret.replace(/&nbsp;/g,' ')
    ret = ret.replace(/<[^>]+>/g,'')
    return ret
  }

  sanitizeHtml(html,all=false,trim=true)
  {
    let ret = ''
    if (all)
      ret = this.sanitizeAllHtml(html)
    else
      ret = this.sanitizeTgHtml(html)
    if (trim)
      ret = ret.trim()
    return ret
  }


}

module.exports = TemplateModule
